# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from typing import Any, Dict, List, Optional

from azure.ai.ml._restclient.v2024_10_01_preview.models import Workspace as RestWorkspace
from azure.ai.ml._restclient.v2024_10_01_preview.models import WorkspaceHubConfig as RestWorkspaceHubConfig
from azure.ai.ml._schema.workspace import HubSchema
from azure.ai.ml.constants._common import WorkspaceKind
from azure.ai.ml.entities._credentials import IdentityConfiguration
from azure.ai.ml.entities._workspace.customer_managed_key import CustomerManagedKey
from azure.ai.ml.entities._workspace.network_acls import NetworkAcls
from azure.ai.ml.entities._workspace.networking import ManagedNetwork
from azure.ai.ml.entities._workspace.workspace import Workspace


class Hub(Workspace):
    """A Hub is a special type of workspace that acts as a parent and resource container for lightweight child
    workspaces called projects. Resources like the hub's storage account, key vault,
    and container registry are shared by all child projects.

    As a type of workspace, hub management is controlled by an MLClient's workspace operations.

    :param name: Name of the hub.
    :type name: str
    :param description: Description of the hub.
    :type description: str
    :param tags: Tags of the hub.
    :type tags: dict
    :param display_name: Display name for the hub. This is non-unique within the resource group.
    :type display_name: str
    :param location: The location to create the hub in.
        If not specified, the same location as the resource group will be used.
    :type location: str
    :param resource_group: Name of resource group to create the hub in.
    :type resource_group: str
    :param managed_network: Hub's Managed Network configuration
    :type managed_network: ~azure.ai.ml.entities.ManagedNetwork
    :param storage_account: The resource ID of an existing storage account to use instead of creating a new one.
    :type storage_account: str
    :param key_vault: The resource ID of an existing key vault to use instead of creating a new one.
    :type key_vault: str
    :param container_registry: The resource ID of an existing container registry
        to use instead of creating a new one.
    :type container_registry: str
    :param customer_managed_key: Key vault details for encrypting data with customer-managed keys.
        If not specified, Microsoft-managed keys will be used by default.
    :type customer_managed_key: ~azure.ai.ml.entities.CustomerManagedKey
    :param image_build_compute: The name of the compute target to use for building environment.
        Docker images with the container registry is behind a VNet.
    :type image_build_compute: str
    :param public_network_access: Whether to allow public endpoint connectivity.
        when a workspace is private link enabled.
    :type public_network_access: str
    :param network_acls: The network access control list (ACL) settings of the workspace.
    :type network_acls: ~azure.ai.ml.entities.NetworkAcls
    :param identity: The hub's Managed Identity (user assigned, or system assigned).
    :type identity: ~azure.ai.ml.entities.IdentityConfiguration
    :param primary_user_assigned_identity: The hub's primary user assigned identity.
    :type primary_user_assigned_identity: str
    :param enable_data_isolation: A flag to determine if workspace has data isolation enabled.
        The flag can only be set at the creation phase, it can't be updated.
    :type enable_data_isolation: bool
    :param default_resource_group: The resource group that will be used by projects
        created under this hub if no resource group is specified.
    :type default_resource_group: str
    :param kwargs: A dictionary of additional configuration parameters.
    :type kwargs: dict

    .. literalinclude:: ../samples/ml_samples_workspace.py
            :start-after: [START workspace_hub]
            :end-before: [END workspace_hub]
            :language: python
            :dedent: 8
            :caption: Creating a Hub object.
    """

    # The field 'additional_workspace_storage_accounts' exists in the API but is currently unused.

    def __init__(
        self,
        *,
        name: str,
        description: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
        display_name: Optional[str] = None,
        location: Optional[str] = None,
        resource_group: Optional[str] = None,
        managed_network: Optional[ManagedNetwork] = None,
        storage_account: Optional[str] = None,
        key_vault: Optional[str] = None,
        container_registry: Optional[str] = None,
        customer_managed_key: Optional[CustomerManagedKey] = None,
        public_network_access: Optional[str] = None,
        network_acls: Optional[NetworkAcls] = None,
        identity: Optional[IdentityConfiguration] = None,
        primary_user_assigned_identity: Optional[str] = None,
        enable_data_isolation: bool = False,
        default_resource_group: Optional[str] = None,
        associated_workspaces: Optional[List[str]] = None,  # hidden input for rest->client conversions.
        **kwargs: Any,
    ):
        self._workspace_id = kwargs.pop("workspace_id", "")
        # Ensure user can't overwrite/double input kind.
        kwargs.pop("kind", None)
        super().__init__(
            name=name,
            description=description,
            tags=tags,
            kind=WorkspaceKind.HUB,
            display_name=display_name,
            location=location,
            storage_account=storage_account,
            key_vault=key_vault,
            container_registry=container_registry,
            resource_group=resource_group,
            customer_managed_key=customer_managed_key,
            public_network_access=public_network_access,
            network_acls=network_acls,
            identity=identity,
            primary_user_assigned_identity=primary_user_assigned_identity,
            managed_network=managed_network,
            enable_data_isolation=enable_data_isolation,
            **kwargs,
        )
        self._default_resource_group = default_resource_group
        self._associated_workspaces = associated_workspaces

    @classmethod
    def _get_schema_class(cls):
        return HubSchema

    @classmethod
    def _from_rest_object(cls, rest_obj: RestWorkspace, v2_service_context: Optional[object] = None) -> Optional["Hub"]:
        if not rest_obj:
            return None

        workspace_object = Workspace._from_rest_object(rest_obj, v2_service_context)

        default_resource_group = None

        if hasattr(rest_obj, "workspace_hub_config"):
            if rest_obj.workspace_hub_config and isinstance(rest_obj.workspace_hub_config, RestWorkspaceHubConfig):
                default_resource_group = rest_obj.workspace_hub_config.default_workspace_resource_group

        if workspace_object is not None:
            return Hub(
                name=workspace_object.name if workspace_object.name is not None else "",
                description=workspace_object.description,
                tags=workspace_object.tags,
                display_name=workspace_object.display_name,
                location=workspace_object.location,
                resource_group=workspace_object.resource_group,
                managed_network=workspace_object.managed_network,
                customer_managed_key=workspace_object.customer_managed_key,
                public_network_access=workspace_object.public_network_access,
                network_acls=workspace_object.network_acls,
                identity=workspace_object.identity,
                primary_user_assigned_identity=workspace_object.primary_user_assigned_identity,
                storage_account=rest_obj.storage_account,
                key_vault=rest_obj.key_vault,
                container_registry=rest_obj.container_registry,
                workspace_id=rest_obj.workspace_id,
                enable_data_isolation=rest_obj.enable_data_isolation,
                default_resource_group=default_resource_group,
                associated_workspaces=rest_obj.associated_workspaces if rest_obj.associated_workspaces else [],
                id=rest_obj.id,
            )
        return None

    # Helper function to deal with sub-rest object conversion.
    def _hub_values_to_rest_object(self) -> RestWorkspaceHubConfig:
        additional_workspace_storage_accounts = None
        default_resource_group = None
        if hasattr(self, "additional_workspace_storage_accounts"):
            additional_workspace_storage_accounts = None
        if hasattr(self, "default_resource_group"):
            default_resource_group = None
        return RestWorkspaceHubConfig(
            additional_workspace_storage_accounts=additional_workspace_storage_accounts,
            default_workspace_resource_group=default_resource_group,
        )

    def _to_rest_object(self) -> RestWorkspace:
        restWorkspace = super()._to_rest_object()
        restWorkspace.workspace_hub_config = self._hub_values_to_rest_object()
        return restWorkspace

    @property
    def default_resource_group(self) -> Optional[str]:
        """The default resource group for this hub and its children.

        :return: The resource group.
        :rtype: Optional[str]
        """
        return self._default_resource_group

    @default_resource_group.setter
    def default_resource_group(self, value: str):
        """Set the default resource group for child projects of this hub.

        :param value: The new resource group.
        :type value: str
        """
        if not value:
            return
        self._default_resource_group = value

    # No setter, read-only
    @property
    def associated_workspaces(self) -> Optional[List[str]]:
        """The workspaces associated with the hub.

        :return: The resource group.
        :rtype:  Optional[List[str]]
        """
        return self._associated_workspaces



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""This package includes the type classes which could be used in dsl.pipeline,
command function, or any other place that requires job inputs/outputs.

.. note::

    The following pseudo-code shows how to create a pipeline with such classes.

    .. code-block:: python

        @pipeline()
        def some_pipeline(
            input_param: Input(type="uri_folder", path="xxx", mode="ro_mount"),
            int_param0: Input(type="integer", default=0, min=-3, max=10),
            int_param1 = 2
            str_param = 'abc',
        ):
            pass


    The following pseudo-code shows how to create a command with such classes.

    .. code-block:: python

        my_command = command(
            name="my_command",
            display_name="my_command",
            description="This is a command",
            tags=dict(),
            command="python train.py --input-data ${{inputs.input_data}} --lr ${{inputs.learning_rate}}",
            code="./src",
            compute="cpu-cluster",
            environment="my-env:1",
            distribution=MpiDistribution(process_count_per_instance=4),
            environment_variables=dict(foo="bar"),
            # Customers can still do this:
            # resources=Resources(instance_count=2, instance_type="STANDARD_D2"),
            # limits=Limits(timeout=300),
            inputs={
                "float": Input(type="number", default=1.1, min=0, max=5),
                "integer": Input(type="integer", default=2, min=-1, max=4),
                "integer1": 2,
                "string0": Input(type="string", default="default_str0"),
                "string1": "default_str1",
                "boolean": Input(type="boolean", default=False),
                "uri_folder": Input(type="uri_folder", path="https://my-blob/path/to/data", mode="ro_mount"),
                "uri_file": Input(type="uri_file", path="https://my-blob/path/to/data", mode="download"),
            },
            outputs={"my_model": Output(type="mlflow_model")},
        )
        node = my_command()
"""

from .enum_input import EnumInput
from .external_data import Database, FileSystem
from .group_input import GroupInput
from .input import Input
from .output import Output
from .utils import _get_param_with_standard_annotation, is_group

__all__ = [
    "Input",
    "Output",
    "EnumInput",
    "GroupInput",
    "is_group",
    "_get_param_with_standard_annotation",
    "Database",
    "FileSystem",
]



import os
import tempfile
from pathlib import Path

import pytest

from azure.ai.ml.operations._local_job_invoker import (
    _get_creationflags_and_startupinfo_for_background_process,
    patch_invocation_script_serialization,
)


@pytest.mark.unittest
@pytest.mark.training_experiences_test
class TestLocalJobInvoker:
    def test_serialize_patch(self):
        in_text = """unread strings
        everythng before are ignored
        continue to ignore --snapshots '"""
        in_text += '[{"Id": "abc-123"}]'
        in_text += """' Everything after is ignored"""

        expected_out = (
            """unread strings
        everythng before are ignored
        continue to ignore --snapshots"""
            + ' "[{\\"Id\\": \\"abc-123\\"}]" '
            + """Everything after is ignored"""
        )

        with tempfile.TemporaryDirectory() as temp_dir:
            with tempfile.NamedTemporaryFile(mode="w", delete=False, dir=temp_dir) as f:
                f.write(in_text)
                f.flush()
                patch_invocation_script_serialization(Path(f.name).resolve())
            assert Path(f.name).read_text() == expected_out

    def test_serialize_patch_no_op(self):
        # Validate that the previously escaped string does not modify if funciton is run
        dummy_file = Path("./dummy_file2.txt").resolve()
        expected_out = (
            """unread strings
        everythng before are ignored
        continue to ignore --snapshots"""
            + ' "[{\\"Id\\": \\"abc-123\\"}]" '
            + """Everything after is ignored"""
        )
        dummy_file.write_text(expected_out)
        # Should change nothing if correctly serialized
        patch_invocation_script_serialization(dummy_file)

        assert dummy_file.read_text() == expected_out
        dummy_file.unlink()

    def test_creation_flags(self):
        flags = _get_creationflags_and_startupinfo_for_background_process()
        if os.name == "nt":
            assert flags["creationflags"] == 16
            assert flags["startupinfo"] is not None
            # After validating Windows flags, rerun with override os set.
            flags = _get_creationflags_and_startupinfo_for_background_process("linux")

        assert flags == {"stderr": -2, "stdin": -3, "stdout": -3}



import os
import argparse
import logging
import sys
import glob


def get_arg_parser(parser=None):
    """Parse the command line arguments for merge using argparse.

    Args:
        parser (argparse.ArgumentParser or CompliantArgumentParser):
        an argument parser instance

    Returns:
        ArgumentParser: the argument parser instance

    Notes:
        if parser is None, creates a new parser instance
    """
    # add arguments that are specific to the component
    if parser is None:
        parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument("--raw_training_data", type=str, required=True, help="")
    parser.add_argument("--raw_testing_data", type=str, required=True, help="")
    parser.add_argument("--train_output", type=str, required=True, help="")
    parser.add_argument("--test_output", type=str, required=True, help="")
    parser.add_argument("--metrics_prefix", type=str, required=False, help="Metrics prefix")
    return parser


def test_input(path):
    file_list = glob.glob(os.path.join(path, "*.*"), recursive=True)
    print(f"Found {len(file_list)} files in {path}")

    print(f"Reading files from {path}")
    for file in file_list:
        print(f" -- Reading {file}")
        with open(file, "r") as f:
            f.read()


def test_output(path):
    with open(os.path.join(path, "output.txt"), "w") as f:
        f.write("Hello World!")


def main(cli_args=None):
    """Component main function.

    It parses arguments and executes run() with the right arguments.

    Args:
        cli_args (List[str], optional): list of args to feed script, useful for debugging. Defaults to None.
    """
    # build an arg parser
    parser = get_arg_parser()

    # run the parser on cli args
    args = parser.parse_args(cli_args)

    print(f"Running script with arguments: {args}")
    test_input(args.raw_training_data)
    test_input(args.raw_testing_data)
    test_output(args.train_output)
    test_output(args.test_output)


if __name__ == "__main__":
    # Set logging to sys.out
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)
    log_format = logging.Formatter("[%(asctime)s] [%(levelname)s] - %(message)s")
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(log_format)
    logger.addHandler(handler)

    main()



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from .automl_nlp_job import AutoMLNLPJob
from .nlp_featurization_settings import NlpFeaturizationSettings
from .nlp_fixed_parameters import NlpFixedParameters
from .nlp_limit_settings import NlpLimitSettings
from .nlp_search_space import NlpSearchSpace
from .nlp_sweep_settings import NlpSweepSettings
from .text_classification_job import TextClassificationJob
from .text_classification_multilabel_job import TextClassificationMultilabelJob
from .text_ner_job import TextNerJob

__all__ = [
    "AutoMLNLPJob",
    "NlpFeaturizationSettings",
    "NlpFixedParameters",
    "NlpLimitSettings",
    "NlpSearchSpace",
    "NlpSweepSettings",
    "TextClassificationJob",
    "TextClassificationMultilabelJob",
    "TextNerJob",
]



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------
import functools
from typing import Any, AsyncIterable, Callable, Dict, Generic, Optional, TypeVar
import warnings

from azure.core.async_paging import AsyncItemPaged, AsyncList
from azure.core.exceptions import ClientAuthenticationError, HttpResponseError, ResourceExistsError, ResourceNotFoundError, map_error
from azure.core.pipeline import PipelineResponse
from azure.core.pipeline.transport import AsyncHttpResponse
from azure.core.rest import HttpRequest
from azure.core.tracing.decorator import distributed_trace
from azure.core.tracing.decorator_async import distributed_trace_async
from azure.mgmt.core.exceptions import ARMErrorFormat

from ... import models as _models
from ..._vendor import _convert_request
from ...operations._data_version_operations import build_batch_get_resolved_uris_request, build_create_request, build_create_unregistered_input_data_request, build_create_unregistered_output_data_request, build_delete_request, build_exists_request, build_get_by_asset_id_request, build_get_request, build_list_request, build_modify_request, build_registered_existing_data_request
T = TypeVar('T')
ClsType = Optional[Callable[[PipelineResponse[HttpRequest, AsyncHttpResponse], T, Dict[str, Any]], Any]]

class DataVersionOperations:
    """DataVersionOperations async operations.

    You should not instantiate this class directly. Instead, you should create a Client instance that
    instantiates it for you and attaches it as an attribute.

    :ivar models: Alias to model classes used in this operation group.
    :type models: ~azure.mgmt.machinelearningservices.models
    :param client: Client for service requests.
    :param config: Configuration of service client.
    :param serializer: An object model serializer.
    :param deserializer: An object model deserializer.
    """

    models = _models

    def __init__(self, client, config, serializer, deserializer) -> None:
        self._client = client
        self._serialize = serializer
        self._deserialize = deserializer
        self._config = config

    @distributed_trace_async
    async def create(
        self,
        subscription_id: str,
        resource_group_name: str,
        workspace_name: str,
        name: str,
        body: Optional["_models.DataVersion"] = None,
        **kwargs: Any
    ) -> "_models.DataVersionEntity":
        """create.

        :param subscription_id: The Azure Subscription ID.
        :type subscription_id: str
        :param resource_group_name: The Name of the resource group in which the workspace is located.
        :type resource_group_name: str
        :param workspace_name: The name of the workspace.
        :type workspace_name: str
        :param name:
        :type name: str
        :param body:
        :type body: ~azure.mgmt.machinelearningservices.models.DataVersion
        :keyword callable cls: A custom type or function that will be passed the direct response
        :return: DataVersionEntity, or the result of cls(response)
        :rtype: ~azure.mgmt.machinelearningservices.models.DataVersionEntity
        :raises: ~azure.core.exceptions.HttpResponseError
        """
        cls = kwargs.pop('cls', None)  # type: ClsType["_models.DataVersionEntity"]
        error_map = {
            401: ClientAuthenticationError, 404: ResourceNotFoundError, 409: ResourceExistsError
        }
        error_map.update(kwargs.pop('error_map', {}))

        content_type = kwargs.pop('content_type', "application/json")  # type: Optional[str]

        if body is not None:
            _json = self._serialize.body(body, 'DataVersion')
        else:
            _json = None

        request = build_create_request(
            subscription_id=subscription_id,
            resource_group_name=resource_group_name,
            workspace_name=workspace_name,
            name=name,
            content_type=content_type,
            json=_json,
            template_url=self.create.metadata['url'],
        )
        request = _convert_request(request)
        request.url = self._client.format_url(request.url)

        pipeline_response = await self._client._pipeline.run(request, stream=False, **kwargs)
        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            error = self._deserialize.failsafe_deserialize(_models.ErrorResponse, pipeline_response)
            raise HttpResponseError(response=response, model=error, error_format=ARMErrorFormat)

        deserialized = self._deserialize('DataVersionEntity', pipeline_response)

        if cls:
            return cls(pipeline_response, deserialized, {})

        return deserialized

    create.metadata = {'url': '/data/v1.0/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.MachineLearningServices/workspaces/{workspaceName}/dataversion/{name}/versions'}  # type: ignore


    @distributed_trace
    def list(
        self,
        subscription_id: str,
        resource_group_name: str,
        workspace_name: str,
        name: str,
        order_by: Optional[str] = None,
        top: Optional[int] = None,
        **kwargs: Any
    ) -> AsyncIterable["_models.PaginatedDataVersionEntityList"]:
        """list.

        :param subscription_id: The Azure Subscription ID.
        :type subscription_id: str
        :param resource_group_name: The Name of the resource group in which the workspace is located.
        :type resource_group_name: str
        :param workspace_name: The name of the workspace.
        :type workspace_name: str
        :param name:
        :type name: str
        :param order_by:
        :type order_by: str
        :param top:
        :type top: int
        :keyword callable cls: A custom type or function that will be passed the direct response
        :return: An iterator like instance of either PaginatedDataVersionEntityList or the result of
         cls(response)
        :rtype:
         ~azure.core.async_paging.AsyncItemPaged[~azure.mgmt.machinelearningservices.models.PaginatedDataVersionEntityList]
        :raises: ~azure.core.exceptions.HttpResponseError
        """
        cls = kwargs.pop('cls', None)  # type: ClsType["_models.PaginatedDataVersionEntityList"]
        error_map = {
            401: ClientAuthenticationError, 404: ResourceNotFoundError, 409: ResourceExistsError
        }
        error_map.update(kwargs.pop('error_map', {}))
        def prepare_request(next_link=None):
            if not next_link:
                
                request = build_list_request(
                    subscription_id=subscription_id,
                    resource_group_name=resource_group_name,
                    workspace_name=workspace_name,
                    name=name,
                    order_by=order_by,
                    top=top,
                    template_url=self.list.metadata['url'],
                )
                request = _convert_request(request)
                request.url = self._client.format_url(request.url)

            else:
                
                request = build_list_request(
                    subscription_id=subscription_id,
                    resource_group_name=resource_group_name,
                    workspace_name=workspace_name,
                    name=name,
                    order_by=order_by,
                    top=top,
                    template_url=next_link,
                )
                request = _convert_request(request)
                request.url = self._client.format_url(request.url)
                request.method = "GET"
            return request

        async def extract_data(pipeline_response):
            deserialized = self._deserialize("PaginatedDataVersionEntityList", pipeline_response)
            list_of_elem = deserialized.value
            if cls:
                list_of_elem = cls(list_of_elem)
            return deserialized.next_link or None, AsyncList(list_of_elem)

        async def get_next(next_link=None):
            request = prepare_request(next_link)

            pipeline_response = await self._client._pipeline.run(request, stream=False, **kwargs)
            response = pipeline_response.http_response

            if response.status_code not in [200]:
                map_error(status_code=response.status_code, response=response, error_map=error_map)
                error = self._deserialize.failsafe_deserialize(_models.ErrorResponse, pipeline_response)
                raise HttpResponseError(response=response, model=error, error_format=ARMErrorFormat)

            return pipeline_response


        return AsyncItemPaged(
            get_next, extract_data
        )
    list.metadata = {'url': '/data/v1.0/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.MachineLearningServices/workspaces/{workspaceName}/dataversion/{name}/versions'}  # type: ignore

    @distributed_trace_async
    async def get(
        self,
        subscription_id: str,
        resource_group_name: str,
        workspace_name: str,
        name: str,
        version: str,
        **kwargs: Any
    ) -> "_models.DataVersionEntity":
        """get.

        :param subscription_id: The Azure Subscription ID.
        :type subscription_id: str
        :param resource_group_name: The Name of the resource group in which the workspace is located.
        :type resource_group_name: str
        :param workspace_name: The name of the workspace.
        :type workspace_name: str
        :param name:
        :type name: str
        :param version:
        :type version: str
        :keyword callable cls: A custom type or function that will be passed the direct response
        :return: DataVersionEntity, or the result of cls(response)
        :rtype: ~azure.mgmt.machinelearningservices.models.DataVersionEntity
        :raises: ~azure.core.exceptions.HttpResponseError
        """
        cls = kwargs.pop('cls', None)  # type: ClsType["_models.DataVersionEntity"]
        error_map = {
            401: ClientAuthenticationError, 404: ResourceNotFoundError, 409: ResourceExistsError
        }
        error_map.update(kwargs.pop('error_map', {}))

        
        request = build_get_request(
            subscription_id=subscription_id,
            resource_group_name=resource_group_name,
            workspace_name=workspace_name,
            name=name,
            version=version,
            template_url=self.get.metadata['url'],
        )
        request = _convert_request(request)
        request.url = self._client.format_url(request.url)

        pipeline_response = await self._client._pipeline.run(request, stream=False, **kwargs)
        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            error = self._deserialize.failsafe_deserialize(_models.ErrorResponse, pipeline_response)
            raise HttpResponseError(response=response, model=error, error_format=ARMErrorFormat)

        deserialized = self._deserialize('DataVersionEntity', pipeline_response)

        if cls:
            return cls(pipeline_response, deserialized, {})

        return deserialized

    get.metadata = {'url': '/data/v1.0/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.MachineLearningServices/workspaces/{workspaceName}/dataversion/{name}/versions/{version}'}  # type: ignore


    @distributed_trace_async
    async def modify(
        self,
        subscription_id: str,
        resource_group_name: str,
        workspace_name: str,
        name: str,
        version: str,
        body: Optional["_models.DataVersionMutable"] = None,
        **kwargs: Any
    ) -> "_models.DataVersionEntity":
        """modify.

        :param subscription_id: The Azure Subscription ID.
        :type subscription_id: str
        :param resource_group_name: The Name of the resource group in which the workspace is located.
        :type resource_group_name: str
        :param workspace_name: The name of the workspace.
        :type workspace_name: str
        :param name:
        :type name: str
        :param version:
        :type version: str
        :param body:
        :type body: ~azure.mgmt.machinelearningservices.models.DataVersionMutable
        :keyword callable cls: A custom type or function that will be passed the direct response
        :return: DataVersionEntity, or the result of cls(response)
        :rtype: ~azure.mgmt.machinelearningservices.models.DataVersionEntity
        :raises: ~azure.core.exceptions.HttpResponseError
        """
        cls = kwargs.pop('cls', None)  # type: ClsType["_models.DataVersionEntity"]
        error_map = {
            401: ClientAuthenticationError, 404: ResourceNotFoundError, 409: ResourceExistsError
        }
        error_map.update(kwargs.pop('error_map', {}))

        content_type = kwargs.pop('content_type', "application/json")  # type: Optional[str]

        if body is not None:
            _json = self._serialize.body(body, 'DataVersionMutable')
        else:
            _json = None

        request = build_modify_request(
            subscription_id=subscription_id,
            resource_group_name=resource_group_name,
            workspace_name=workspace_name,
            name=name,
            version=version,
            content_type=content_type,
            json=_json,
            template_url=self.modify.metadata['url'],
        )
        request = _convert_request(request)
        request.url = self._client.format_url(request.url)

        pipeline_response = await self._client._pipeline.run(request, stream=False, **kwargs)
        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            error = self._deserialize.failsafe_deserialize(_models.ErrorResponse, pipeline_response)
            raise HttpResponseError(response=response, model=error, error_format=ARMErrorFormat)

        deserialized = self._deserialize('DataVersionEntity', pipeline_response)

        if cls:
            return cls(pipeline_response, deserialized, {})

        return deserialized

    modify.metadata = {'url': '/data/v1.0/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.MachineLearningServices/workspaces/{workspaceName}/dataversion/{name}/versions/{version}'}  # type: ignore


    @distributed_trace_async
    async def delete(
        self,
        subscription_id: str,
        resource_group_name: str,
        workspace_name: str,
        name: str,
        version: str,
        **kwargs: Any
    ) -> "_models.HttpResponseMessage":
        """delete.

        :param subscription_id: The Azure Subscription ID.
        :type subscription_id: str
        :param resource_group_name: The Name of the resource group in which the workspace is located.
        :type resource_group_name: str
        :param workspace_name: The name of the workspace.
        :type workspace_name: str
        :param name:
        :type name: str
        :param version:
        :type version: str
        :keyword callable cls: A custom type or function that will be passed the direct response
        :return: HttpResponseMessage, or the result of cls(response)
        :rtype: ~azure.mgmt.machinelearningservices.models.HttpResponseMessage
        :raises: ~azure.core.exceptions.HttpResponseError
        """
        cls = kwargs.pop('cls', None)  # type: ClsType["_models.HttpResponseMessage"]
        error_map = {
            401: ClientAuthenticationError, 404: ResourceNotFoundError, 409: ResourceExistsError
        }
        error_map.update(kwargs.pop('error_map', {}))

        
        request = build_delete_request(
            subscription_id=subscription_id,
            resource_group_name=resource_group_name,
            workspace_name=workspace_name,
            name=name,
            version=version,
            template_url=self.delete.metadata['url'],
        )
        request = _convert_request(request)
        request.url = self._client.format_url(request.url)

        pipeline_response = await self._client._pipeline.run(request, stream=False, **kwargs)
        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            error = self._deserialize.failsafe_deserialize(_models.ErrorResponse, pipeline_response)
            raise HttpResponseError(response=response, model=error, error_format=ARMErrorFormat)

        deserialized = self._deserialize('HttpResponseMessage', pipeline_response)

        if cls:
            return cls(pipeline_response, deserialized, {})

        return deserialized

    delete.metadata = {'url': '/data/v1.0/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.MachineLearningServices/workspaces/{workspaceName}/dataversion/{name}/versions/{version}'}  # type: ignore


    @distributed_trace_async
    async def exists(
        self,
        subscription_id: str,
        resource_group_name: str,
        workspace_name: str,
        name: str,
        version: str,
        **kwargs: Any
    ) -> bool:
        """exists.

        :param subscription_id: The Azure Subscription ID.
        :type subscription_id: str
        :param resource_group_name: The Name of the resource group in which the workspace is located.
        :type resource_group_name: str
        :param workspace_name: The name of the workspace.
        :type workspace_name: str
        :param name:
        :type name: str
        :param version:
        :type version: str
        :keyword callable cls: A custom type or function that will be passed the direct response
        :return: bool, or the result of cls(response)
        :rtype: bool
        :raises: ~azure.core.exceptions.HttpResponseError
        """
        cls = kwargs.pop('cls', None)  # type: ClsType[bool]
        error_map = {
            401: ClientAuthenticationError, 404: ResourceNotFoundError, 409: ResourceExistsError
        }
        error_map.update(kwargs.pop('error_map', {}))

        
        request = build_exists_request(
            subscription_id=subscription_id,
            resource_group_name=resource_group_name,
            workspace_name=workspace_name,
            name=name,
            version=version,
            template_url=self.exists.metadata['url'],
        )
        request = _convert_request(request)
        request.url = self._client.format_url(request.url)

        pipeline_response = await self._client._pipeline.run(request, stream=False, **kwargs)
        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            error = self._deserialize.failsafe_deserialize(_models.ErrorResponse, pipeline_response)
            raise HttpResponseError(response=response, model=error, error_format=ARMErrorFormat)

        deserialized = self._deserialize('bool', pipeline_response)

        if cls:
            return cls(pipeline_response, deserialized, {})

        return deserialized

    exists.metadata = {'url': '/data/v1.0/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.MachineLearningServices/workspaces/{workspaceName}/dataversion/{name}/versions/{version}/exists'}  # type: ignore


    @distributed_trace_async
    async def get_by_asset_id(
        self,
        subscription_id: str,
        resource_group_name: str,
        workspace_name: str,
        body: Optional["_models.AssetId"] = None,
        **kwargs: Any
    ) -> "_models.DataVersionEntity":
        """get_by_asset_id.

        :param subscription_id: The Azure Subscription ID.
        :type subscription_id: str
        :param resource_group_name: The Name of the resource group in which the workspace is located.
        :type resource_group_name: str
        :param workspace_name: The name of the workspace.
        :type workspace_name: str
        :param body:
        :type body: ~azure.mgmt.machinelearningservices.models.AssetId
        :keyword callable cls: A custom type or function that will be passed the direct response
        :return: DataVersionEntity, or the result of cls(response)
        :rtype: ~azure.mgmt.machinelearningservices.models.DataVersionEntity
        :raises: ~azure.core.exceptions.HttpResponseError
        """
        cls = kwargs.pop('cls', None)  # type: ClsType["_models.DataVersionEntity"]
        error_map = {
            401: ClientAuthenticationError, 404: ResourceNotFoundError, 409: ResourceExistsError
        }
        error_map.update(kwargs.pop('error_map', {}))

        content_type = kwargs.pop('content_type', "application/json")  # type: Optional[str]

        if body is not None:
            _json = self._serialize.body(body, 'AssetId')
        else:
            _json = None

        request = build_get_by_asset_id_request(
            subscription_id=subscription_id,
            resource_group_name=resource_group_name,
            workspace_name=workspace_name,
            content_type=content_type,
            json=_json,
            template_url=self.get_by_asset_id.metadata['url'],
        )
        request = _convert_request(request)
        request.url = self._client.format_url(request.url)

        pipeline_response = await self._client._pipeline.run(request, stream=False, **kwargs)
        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            error = self._deserialize.failsafe_deserialize(_models.ErrorResponse, pipeline_response)
            raise HttpResponseError(response=response, model=error, error_format=ARMErrorFormat)

        deserialized = self._deserialize('DataVersionEntity', pipeline_response)

        if cls:
            return cls(pipeline_response, deserialized, {})

        return deserialized

    get_by_asset_id.metadata = {'url': '/data/v1.0/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.MachineLearningServices/workspaces/{workspaceName}/dataversion/getByAssetId'}  # type: ignore


    @distributed_trace_async
    async def create_unregistered_input_data(
        self,
        subscription_id: str,
        resource_group_name: str,
        workspace_name: str,
        body: Optional["_models.CreateUnregisteredInputData"] = None,
        **kwargs: Any
    ) -> "_models.DataContainerEntity":
        """create_unregistered_input_data.

        :param subscription_id: The Azure Subscription ID.
        :type subscription_id: str
        :param resource_group_name: The Name of the resource group in which the workspace is located.
        :type resource_group_name: str
        :param workspace_name: The name of the workspace.
        :type workspace_name: str
        :param body:
        :type body: ~azure.mgmt.machinelearningservices.models.CreateUnregisteredInputData
        :keyword callable cls: A custom type or function that will be passed the direct response
        :return: DataContainerEntity, or the result of cls(response)
        :rtype: ~azure.mgmt.machinelearningservices.models.DataContainerEntity
        :raises: ~azure.core.exceptions.HttpResponseError
        """
        cls = kwargs.pop('cls', None)  # type: ClsType["_models.DataContainerEntity"]
        error_map = {
            401: ClientAuthenticationError, 404: ResourceNotFoundError, 409: ResourceExistsError
        }
        error_map.update(kwargs.pop('error_map', {}))

        content_type = kwargs.pop('content_type', "application/json")  # type: Optional[str]

        if body is not None:
            _json = self._serialize.body(body, 'CreateUnregisteredInputData')
        else:
            _json = None

        request = build_create_unregistered_input_data_request(
            subscription_id=subscription_id,
            resource_group_name=resource_group_name,
            workspace_name=workspace_name,
            content_type=content_type,
            json=_json,
            template_url=self.create_unregistered_input_data.metadata['url'],
        )
        request = _convert_request(request)
        request.url = self._client.format_url(request.url)

        pipeline_response = await self._client._pipeline.run(request, stream=False, **kwargs)
        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            error = self._deserialize.failsafe_deserialize(_models.ErrorResponse, pipeline_response)
            raise HttpResponseError(response=response, model=error, error_format=ARMErrorFormat)

        deserialized = self._deserialize('DataContainerEntity', pipeline_response)

        if cls:
            return cls(pipeline_response, deserialized, {})

        return deserialized

    create_unregistered_input_data.metadata = {'url': '/data/v1.0/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.MachineLearningServices/workspaces/{workspaceName}/dataversion/createUnregisteredInput'}  # type: ignore


    @distributed_trace_async
    async def create_unregistered_output_data(
        self,
        subscription_id: str,
        resource_group_name: str,
        workspace_name: str,
        body: Optional["_models.CreateUnregisteredOutputData"] = None,
        **kwargs: Any
    ) -> "_models.DataContainerEntity":
        """create_unregistered_output_data.

        :param subscription_id: The Azure Subscription ID.
        :type subscription_id: str
        :param resource_group_name: The Name of the resource group in which the workspace is located.
        :type resource_group_name: str
        :param workspace_name: The name of the workspace.
        :type workspace_name: str
        :param body:
        :type body: ~azure.mgmt.machinelearningservices.models.CreateUnregisteredOutputData
        :keyword callable cls: A custom type or function that will be passed the direct response
        :return: DataContainerEntity, or the result of cls(response)
        :rtype: ~azure.mgmt.machinelearningservices.models.DataContainerEntity
        :raises: ~azure.core.exceptions.HttpResponseError
        """
        cls = kwargs.pop('cls', None)  # type: ClsType["_models.DataContainerEntity"]
        error_map = {
            401: ClientAuthenticationError, 404: ResourceNotFoundError, 409: ResourceExistsError
        }
        error_map.update(kwargs.pop('error_map', {}))

        content_type = kwargs.pop('content_type', "application/json")  # type: Optional[str]

        if body is not None:
            _json = self._serialize.body(body, 'CreateUnregisteredOutputData')
        else:
            _json = None

        request = build_create_unregistered_output_data_request(
            subscription_id=subscription_id,
            resource_group_name=resource_group_name,
            workspace_name=workspace_name,
            content_type=content_type,
            json=_json,
            template_url=self.create_unregistered_output_data.metadata['url'],
        )
        request = _convert_request(request)
        request.url = self._client.format_url(request.url)

        pipeline_response = await self._client._pipeline.run(request, stream=False, **kwargs)
        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            error = self._deserialize.failsafe_deserialize(_models.ErrorResponse, pipeline_response)
            raise HttpResponseError(response=response, model=error, error_format=ARMErrorFormat)

        deserialized = self._deserialize('DataContainerEntity', pipeline_response)

        if cls:
            return cls(pipeline_response, deserialized, {})

        return deserialized

    create_unregistered_output_data.metadata = {'url': '/data/v1.0/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.MachineLearningServices/workspaces/{workspaceName}/dataversion/createUnregisteredOutput'}  # type: ignore


    @distributed_trace_async
    async def registered_existing_data(
        self,
        subscription_id: str,
        resource_group_name: str,
        workspace_name: str,
        body: Optional["_models.RegisterExistingData"] = None,
        **kwargs: Any
    ) -> "_models.DataContainerEntity":
        """registered_existing_data.

        :param subscription_id: The Azure Subscription ID.
        :type subscription_id: str
        :param resource_group_name: The Name of the resource group in which the workspace is located.
        :type resource_group_name: str
        :param workspace_name: The name of the workspace.
        :type workspace_name: str
        :param body:
        :type body: ~azure.mgmt.machinelearningservices.models.RegisterExistingData
        :keyword callable cls: A custom type or function that will be passed the direct response
        :return: DataContainerEntity, or the result of cls(response)
        :rtype: ~azure.mgmt.machinelearningservices.models.DataContainerEntity
        :raises: ~azure.core.exceptions.HttpResponseError
        """
        cls = kwargs.pop('cls', None)  # type: ClsType["_models.DataContainerEntity"]
        error_map = {
            401: ClientAuthenticationError, 404: ResourceNotFoundError, 409: ResourceExistsError
        }
        error_map.update(kwargs.pop('error_map', {}))

        content_type = kwargs.pop('content_type', "application/json")  # type: Optional[str]

        if body is not None:
            _json = self._serialize.body(body, 'RegisterExistingData')
        else:
            _json = None

        request = build_registered_existing_data_request(
            subscription_id=subscription_id,
            resource_group_name=resource_group_name,
            workspace_name=workspace_name,
            content_type=content_type,
            json=_json,
            template_url=self.registered_existing_data.metadata['url'],
        )
        request = _convert_request(request)
        request.url = self._client.format_url(request.url)

        pipeline_response = await self._client._pipeline.run(request, stream=False, **kwargs)
        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            error = self._deserialize.failsafe_deserialize(_models.ErrorResponse, pipeline_response)
            raise HttpResponseError(response=response, model=error, error_format=ARMErrorFormat)

        deserialized = self._deserialize('DataContainerEntity', pipeline_response)

        if cls:
            return cls(pipeline_response, deserialized, {})

        return deserialized

    registered_existing_data.metadata = {'url': '/data/v1.0/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.MachineLearningServices/workspaces/{workspaceName}/dataversion/registerExisting'}  # type: ignore


    @distributed_trace_async
    async def batch_get_resolved_uris(
        self,
        subscription_id: str,
        resource_group_name: str,
        workspace_name: str,
        body: Optional["_models.BatchGetResolvedURIs"] = None,
        **kwargs: Any
    ) -> "_models.BatchDataUriResponse":
        """batch_get_resolved_uris.

        :param subscription_id: The Azure Subscription ID.
        :type subscription_id: str
        :param resource_group_name: The Name of the resource group in which the workspace is located.
        :type resource_group_name: str
        :param workspace_name: The name of the workspace.
        :type workspace_name: str
        :param body:
        :type body: ~azure.mgmt.machinelearningservices.models.BatchGetResolvedURIs
        :keyword callable cls: A custom type or function that will be passed the direct response
        :return: BatchDataUriResponse, or the result of cls(response)
        :rtype: ~azure.mgmt.machinelearningservices.models.BatchDataUriResponse
        :raises: ~azure.core.exceptions.HttpResponseError
        """
        cls = kwargs.pop('cls', None)  # type: ClsType["_models.BatchDataUriResponse"]
        error_map = {
            401: ClientAuthenticationError, 404: ResourceNotFoundError, 409: ResourceExistsError
        }
        error_map.update(kwargs.pop('error_map', {}))

        content_type = kwargs.pop('content_type', "application/json")  # type: Optional[str]

        if body is not None:
            _json = self._serialize.body(body, 'BatchGetResolvedURIs')
        else:
            _json = None

        request = build_batch_get_resolved_uris_request(
            subscription_id=subscription_id,
            resource_group_name=resource_group_name,
            workspace_name=workspace_name,
            content_type=content_type,
            json=_json,
            template_url=self.batch_get_resolved_uris.metadata['url'],
        )
        request = _convert_request(request)
        request.url = self._client.format_url(request.url)

        pipeline_response = await self._client._pipeline.run(request, stream=False, **kwargs)
        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            error = self._deserialize.failsafe_deserialize(_models.ErrorResponse, pipeline_response)
            raise HttpResponseError(response=response, model=error, error_format=ARMErrorFormat)

        deserialized = self._deserialize('BatchDataUriResponse', pipeline_response)

        if cls:
            return cls(pipeline_response, deserialized, {})

        return deserialized

    batch_get_resolved_uris.metadata = {'url': '/data/v1.0/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.MachineLearningServices/workspaces/{workspaceName}/dataversion/batchGetResolvedUris'}  # type: ignore




from pathlib import Path

import pytest
from marshmallow import ValidationError
from test_utilities.utils import omit_with_wildcard

from azure.ai.ml import Input, load_component
from azure.ai.ml.constants._component import ComponentSource
from azure.ai.ml.dsl import pipeline
from azure.ai.ml.dsl._condition import condition
from azure.ai.ml.dsl._do_while import do_while
from azure.ai.ml.dsl._group_decorator import group
from azure.ai.ml.dsl._parallel_for import parallel_for
from azure.ai.ml.entities._builders.parallel_for import ParallelFor
from azure.ai.ml.entities._job.pipeline._io import InputOutputBase, PipelineInput
from azure.ai.ml.exceptions import ValidationException

from .._util import _DSL_TIMEOUT_SECOND


@pytest.mark.usefixtures(
    "enable_pipeline_private_preview_features",
    "enable_private_preview_schema_features",
    "enable_private_preview_pipeline_node_types",
)
@pytest.mark.timeout(_DSL_TIMEOUT_SECOND)
@pytest.mark.unittest
@pytest.mark.pipeline_test
class TestControlFlowPipelineUT:
    pass


class CustomizedObject:
    pass


class TestIfElseUT(TestControlFlowPipelineUT):
    def test_multiblock_if_else(self):
        hello_world_component_no_paths = load_component(
            source="./tests/test_configs/components/helloworld_component_no_paths.yml"
        )
        basic_component = load_component(
            source="./tests/test_configs/components/component_with_conditional_output/spec.yaml"
        )

        @pipeline()
        def condition_pipeline():
            result = basic_component()
            node1 = hello_world_component_no_paths(component_in_number=1)
            node2 = hello_world_component_no_paths(component_in_number=2)
            condition(condition=result.outputs.output, false_block=[node1, node2])

        pipeline_job = condition_pipeline()
        rest_pipeline_job = pipeline_job._to_rest_object().as_dict()
        assert rest_pipeline_job["properties"]["jobs"]["conditionnode"] == {
            "_source": "DSL",
            "condition": "${{parent.jobs.result.outputs.output}}",
            "false_block": ["${{parent.jobs.node1}}", "${{parent.jobs.node2}}"],
            "type": "if_else",
        }

    def test_if_else_validate(self):
        hello_world_component_no_paths = load_component(
            source="./tests/test_configs/components/helloworld_component_no_paths.yml"
        )
        basic_component = load_component(
            source="./tests/test_configs/components/component_with_conditional_output/spec.yaml"
        )

        @pipeline(compute="cpu-cluster")
        def condition_pipeline():
            result = basic_component()
            node1 = hello_world_component_no_paths(component_in_number=1)
            node2 = hello_world_component_no_paths(component_in_number=2)
            # true block and false block has intersection
            condition(condition=result.outputs.output, false_block=[node1, node2], true_block=[node1])

        with pytest.raises(ValidationError) as e:
            pipeline_job = condition_pipeline()
            pipeline_job._validate(raise_error=True)
        assert "True block and false block cannot contain same nodes: {'${{parent.jobs.node1}}'" in str(e.value)

        @pipeline(compute="cpu-cluster")
        def condition_pipeline():
            result = basic_component()
            node1 = hello_world_component_no_paths()
            node2 = hello_world_component_no_paths()
            # true block and false block has intersection
            condition(condition=result.outputs.output, false_block=[node1], true_block=[node2])

        # no error raise
        pipeline_job = condition_pipeline()
        pipeline_job._validate(raise_error=True)

    def test_create_dsl_condition_illegal_cases(self):
        basic_component = load_component(
            source="./tests/test_configs/components/component_with_conditional_output/spec.yaml"
        )
        basic_node = basic_component()

        with pytest.raises(ValidationException) as e:
            node = condition(condition=1, true_block=basic_node)
            node._validate(raise_error=True)

        assert f"must be an instance of {str}, {bool} or {InputOutputBase}" in str(e.value)

        node = condition(condition=basic_node.outputs.output3, true_block=basic_node)
        node._validate(raise_error=True)

        with pytest.raises(ValidationError) as e:
            node = condition(condition="${{parent.jobs.xxx.outputs.output}}")
            node._validate(raise_error=True)

        assert "True block and false block cannot be empty at the same time." in str(e.value)

        with pytest.raises(ValidationError) as e:
            node = condition(
                condition="${{parent.jobs.xxx.outputs.output}}", true_block=basic_node, false_block=basic_node
            )
            node._validate(raise_error=True)

        assert "True block and false block cannot contain same nodes" in str(e.value)

        with pytest.raises(ValidationException) as e:
            node = condition(condition="xxx", true_block=basic_node)
            node._validate(raise_error=True)

        assert "condition has invalid binding expression: xxx" in str(e.value)

    def test_condition_node(self):
        hello_world_component_no_paths = load_component(
            source="./tests/test_configs/components/helloworld_component_no_paths.yml"
        )
        node1 = hello_world_component_no_paths()
        node1.name = "node1"
        node2 = hello_world_component_no_paths()
        node2.name = "node2"
        control_node = condition(
            condition="${{parent.jobs.condition_predicate.outputs.output}}", false_block=node1, true_block=node2
        )
        assert control_node._to_rest_object() == {
            "_source": "DSL",
            "condition": "${{parent.jobs.condition_predicate.outputs.output}}",
            "false_block": "${{parent.jobs.node1}}",
            "true_block": "${{parent.jobs.node2}}",
            "type": "if_else",
        }

        # test boolean type condition
        control_node = condition(condition=True, false_block=node1, true_block=node2)
        assert control_node._to_rest_object() == {
            "_source": "DSL",
            "condition": True,
            "false_block": "${{parent.jobs.node1}}",
            "true_block": "${{parent.jobs.node2}}",
            "type": "if_else",
        }

    def test_condition_pipeline(self):
        basic_component = load_component(
            source="./tests/test_configs/components/component_with_conditional_output/spec.yaml"
        )

        hello_world_component_no_paths = load_component(
            source="./tests/test_configs/components/helloworld_component_no_paths.yml"
        )

        @pipeline(
            name="test_mldesigner_component_with_conditional_output",
            compute="cpu-cluster",
        )
        def condition_pipeline():
            result = basic_component()
            node1 = hello_world_component_no_paths(component_in_number=1)
            node2 = hello_world_component_no_paths(component_in_number=2)
            condition(condition=result.outputs.output, false_block=node1, true_block=node2)

        pipeline_job = condition_pipeline()
        omit_fields = [
            "name",
            "properties.display_name",
            "properties.jobs.*.componentId",
            "properties.settings",
        ]
        dsl_pipeline_job_dict = omit_with_wildcard(pipeline_job._to_rest_object().as_dict(), *omit_fields)
        assert dsl_pipeline_job_dict["properties"]["jobs"] == {
            "conditionnode": {
                "_source": "DSL",
                "condition": "${{parent.jobs.result.outputs.output}}",
                "false_block": "${{parent.jobs.node1}}",
                "true_block": "${{parent.jobs.node2}}",
                "type": "if_else",
            },
            "node1": {
                "_source": "YAML.COMPONENT",
                "inputs": {"component_in_number": {"job_input_type": "literal", "value": "1"}},
                "name": "node1",
                "type": "command",
            },
            "node2": {
                "_source": "YAML.COMPONENT",
                "inputs": {"component_in_number": {"job_input_type": "literal", "value": "2"}},
                "name": "node2",
                "type": "command",
            },
            "result": {"_source": "YAML.COMPONENT", "name": "result", "type": "command"},
        }

    def test_condition_with_group_input(self):
        hello_world_component_no_paths = load_component(
            source=r"./tests/test_configs/components/helloworld_component_no_paths.yml"
        )

        @group
        class SubGroup:
            num: int

        @group
        class ParentGroup:
            input_group: SubGroup

        @pipeline(
            compute="cpu-cluster",
        )
        def condition_pipeline(group_input: ParentGroup):
            node1 = hello_world_component_no_paths(component_in_number=1)
            condition(condition=group_input.input_group.num < 100, true_block=node1)

        pipeline_job = condition_pipeline(group_input=ParentGroup(input_group=SubGroup(num=10)))
        omit_fields = [
            "name",
            "properties.display_name",
            "properties.jobs.*.componentId",
            "properties.settings",
        ]
        dsl_pipeline_job_dict = omit_with_wildcard(pipeline_job._to_rest_object().as_dict(), *omit_fields)
        assert dsl_pipeline_job_dict["properties"]["jobs"]["expression_component"] == {
            "environment_variables": {"AZURE_ML_CLI_PRIVATE_FEATURES_ENABLED": "true"},
            "name": "expression_component",
            "type": "command",
            "inputs": {"num": {"job_input_type": "literal", "value": "${{parent.inputs.group_input.input_group.num}}"}},
            "_source": "YAML.COMPONENT",
        }


class TestDoWhilePipelineUT(TestControlFlowPipelineUT):
    def test_invalid_do_while_pipeline(self):
        basic_component_func = load_component(
            "./tests/test_configs/dsl_pipeline/do_while/basic_component/component.yml"
        )

        @pipeline
        def do_while_body_pipeline_component(
            component_in_number: Input(type="integer", optional=True),
            component_in_number_1: Input(type="integer", optional=True),
            component_in_path: Input(type="uri_folder"),
        ):
            """E2E dummy train-score-eval pipeline with components defined via yaml."""
            # Call component obj as function: apply given inputs & parameters to create a node in pipeline
            train_with_sample_data = basic_component_func(
                component_in_number=component_in_number,
                component_in_number_1=component_in_number_1,
                component_in_path=component_in_path,
            )
            return train_with_sample_data.outputs

        @pipeline
        def pipeline_with_do_while(
            component_in_number: Input(type="integer"),
            component_in_path: Input(type="uri_folder"),
        ):
            # Create two training pipeline component with different learning rate
            do_while_body_pipeline_1 = do_while_body_pipeline_component(
                component_in_number=component_in_number,
                component_in_number_1=component_in_number,
                component_in_path=component_in_path,
            )
            empty_mapping = do_while(  # noqa: F841
                body=do_while_body_pipeline_1,
                condition=do_while_body_pipeline_1.outputs.is_number_larger_than_zero,
                mapping=None,
                max_iteration_count=5,
            )

            do_while_body_pipeline_2 = do_while_body_pipeline_component(  # noqa: F841
                component_in_number=component_in_number,
                component_in_number_1=component_in_number,
                component_in_path=component_in_path,
            )
            out_of_max_iteration_count_range = do_while(  # noqa: F841
                body=do_while_body_pipeline_2,
                condition=do_while_body_pipeline_2.outputs.is_number_larger_than_zero,
                mapping={
                    do_while_body_pipeline_2.outputs.output_in_number: [
                        do_while_body_pipeline_2.inputs.component_in_number,
                        do_while_body_pipeline_2.inputs.component_in_number_1,
                    ],
                    do_while_body_pipeline_2.outputs.output_in_path: do_while_body_pipeline_2.inputs.component_in_path,
                },
                max_iteration_count=1001,
            )

            body_in_other_loop = do_while(  # noqa: F841
                body=do_while_body_pipeline_1,
                condition=do_while_body_pipeline_1.outputs.is_number_larger_than_zero,
                mapping={
                    do_while_body_pipeline_1.outputs.output_in_number: [
                        do_while_body_pipeline_1.inputs.component_in_number,
                        do_while_body_pipeline_1.inputs.component_in_number_1,
                    ],
                    do_while_body_pipeline_1.outputs.output_in_path: do_while_body_pipeline_1.inputs.component_in_path,
                },
                max_iteration_count=5,
            )

            do_while_body_pipeline_3 = do_while_body_pipeline_component(
                component_in_number=component_in_number,
                component_in_number_1=component_in_number,
                component_in_path=component_in_path,
            )
            invalid_mapping = do_while(  # noqa: F841
                body=do_while_body_pipeline_3,
                condition=do_while_body_pipeline_1.outputs.is_number_larger_than_zero,
                mapping={
                    do_while_body_pipeline_3.outputs.output_in_number: [
                        do_while_body_pipeline_3.inputs.component_in_number,
                        do_while_body_pipeline_1.inputs.component_in_number_1,
                    ],
                    do_while_body_pipeline_1.outputs.output_in_path: do_while_body_pipeline_1.outputs.output_in_path,
                },
                max_iteration_count=5,
            )

            do_while_body_pipeline_4 = do_while_body_pipeline_component(
                component_in_number=component_in_number,
                component_in_number_1=component_in_number,
                component_in_path=component_in_path,
            )
            invalid_condition = do_while(  # noqa: F841
                body=do_while_body_pipeline_4,
                condition=do_while_body_pipeline_4.outputs.output_in_path,
                mapping={
                    do_while_body_pipeline_4.outputs.output_in_number: [
                        do_while_body_pipeline_4.inputs.component_in_number,
                        do_while_body_pipeline_4.inputs.component_in_number_1,
                    ]
                },
                max_iteration_count=5,
            )

        pipeline_job = pipeline_with_do_while(
            component_in_number=2,
            component_in_path=Input(type="uri_folder", path=str(Path(__file__).parent)),
        )
        # set pipeline level compute
        pipeline_job.settings.default_compute = "cpu-cluster"
        result = pipeline_job._customized_validate()
        validate_errors = result._to_dict()["errors"]

        def validate_error_message(expect_errors, validate_errors):
            for path, msg in expect_errors.items():
                acture_errors = list(filter(lambda error: error["path"] == path, validate_errors))
                assert acture_errors, f"Cannot find the error of {path} in validation results."
                for acture_error in acture_errors:
                    assert acture_error["message"] in msg

        expect_errors = {
            "jobs.empty_mapping.mapping": ["Missing data for required field."],
            "jobs.out_of_max_iteration_count_range.limit.max_iteration_count": [
                "The max iteration count cannot be less than 0 or larger than 1000."
            ],
            "jobs.body_in_other_loop.body": ["The body of loop node cannot be promoted as another loop again."],
            "jobs.invalid_mapping.condition": [
                "is_number_larger_than_zero is the output of do_while_body_pipeline_1, dowhile only accept output of the body: do_while_body_pipeline_3."
            ],
            "jobs.invalid_mapping.mapping": [
                "component_in_number_1 is the input of do_while_body_pipeline_1, dowhile only accept input of the body: do_while_body_pipeline_3.",
                "output_in_path is the output of do_while_body_pipeline_1, dowhile only accept output of the body: do_while_body_pipeline_3.",
            ],
            "jobs.invalid_condition.condition": [
                "output_in_path is not a control output and is not primitive type. The condition of dowhile must be the control output or primitive type of the body."
            ],
        }
        validate_error_message(expect_errors, validate_errors)

    def test_infer_dynamic_input_type_from_mapping(self):
        # Pass None to dynamic input in do-while loop body, and provide it in mapping for next iteration,
        # which is a valid case in federated learning.
        from test_configs.dsl_pipeline.dynamic_input_do_while.pipeline import pipeline_job

        assert pipeline_job._customized_validate().passed
        # assert input type of loop body, should both have type now
        assert pipeline_job.jobs["loop_body"].inputs.required_input.type == "uri_folder"
        assert pipeline_job.jobs["loop_body"].inputs.optional_input.type == "uri_folder"


class TestParallelForPipelineUT(TestControlFlowPipelineUT):
    # Need to disable internal components or internal nodes will be in allowed node types
    @pytest.mark.usefixtures("disable_internal_components")
    def test_dsl_parallel_for_pipeline_illegal_cases(self):
        # body unsupported
        parallel_component = load_component(
            source="./tests/test_configs/dsl_pipeline/parallel_component_with_file_input/score.yml"
        )

        basic_component = load_component(
            source="./tests/test_configs/components/component_with_conditional_output/spec.yaml"
        )

        @pipeline
        def invalid_pipeline(test_path1, test_path2):
            body = parallel_component()
            parallel_for(
                body=body,
                items=[{"job_data_path": test_path1}, {"job_data_path": test_path2}],
            )

        with pytest.raises(ValidationException) as e:
            invalid_pipeline(test_path1=Input(path="test_path1"), test_path2=Input(path="test_path2"))
        assert (
            "Expecting (<class 'azure.ai.ml.entities._builders.command.Command'>, "
            "<class 'azure.ai.ml.entities._builders.pipeline.Pipeline'>) for body" in str(e.value)
        )

        # items with invalid type

        @pipeline
        def invalid_pipeline():
            body = basic_component()
            parallel_for(
                body=body,
                items=1,
            )

        with pytest.raises(ValidationException) as e:
            invalid_pipeline()

        assert "got <class 'int'> instead." in str(e.value)

    @pytest.mark.parametrize(
        "items, error_message",
        [
            (
                # items with invalid content type
                {"a": 1},
                "but got <class 'int'> for 1.",
            ),
            (
                # items with empty dict as content
                [],
                "Items is an empty list/dict",
            ),
            (
                # item meta not match
                [{"component_in_path": "test_path1"}, {"component_in_path": "test_path2", "component_in_number": 1}],
                "Items should have same keys with body inputs, but got ",
            ),
            (
                # item inputs not exist in body
                [
                    {"job_data_path": "test_path1"},
                ],
                "Item {'job_data_path': 'test_path1'} got unmatched inputs ",
            ),
            (
                # invalid JSON string items
                '[{"component_in_number": 1}, {}]',
                "Items should have same keys with body inputs",
            ),
            (
                # unsupported item value type
                [
                    {"component_in_number": CustomizedObject()},
                ],
                "Unsupported type",
            ),
            (
                # local file input
                [{"component_in_path": Input(path="./tests/test_configs/components/helloworld_component.yml")}],
                "Local file input",
            ),
            (
                # empty path
                [{"component_in_path": Input(path=None)}],
                "Input path not provided",
            ),
            (
                # dict Input
                [{"component_in_path": {"job_input_path": "azureml://path/to/file"}}],
                "Unsupported type",
            ),
        ],
    )
    def test_dsl_parallel_for_pipeline_illegal_items_content(self, items, error_message):
        basic_component = load_component(source="./tests/test_configs/components/helloworld_component.yml")

        @pipeline
        def invalid_pipeline():
            body = basic_component()
            parallel_for(
                body=body,
                items=items,
            )

        with pytest.raises(ValidationException) as e:
            pipeline_job = invalid_pipeline()
            pipeline_job._validate(raise_error=True)
        assert error_message in str(e.value)

    @pytest.mark.parametrize(
        "items",
        (
            # items with empty dict as content
            [{}],
        ),
    )
    def test_dsl_parallel_for_pipeline_legal_items_content(self, items):
        basic_component = load_component(source="./tests/test_configs/components/helloworld_component.yml")

        @pipeline
        def valid_pipeline():
            body = basic_component()
            parallel_for(
                body=body,
                items=items,
            )

        valid_pipeline()

    def test_dsl_parallel_for_pipeline_items(self):
        # TODO: submit those pipelines

        basic_component = load_component(source="./tests/test_configs/components/helloworld_component.yml")
        complex_component = load_component(source="./tests/test_configs/components/input_types_component.yml")

        # binding in items

        @pipeline
        def my_pipeline(test_path1, test_path2):
            body = basic_component()
            parallel_for(
                body=body,
                items=[{"component_in_path": test_path1}, {"component_in_path": test_path2}],
            )

        my_job = my_pipeline(test_path1=Input(path="test_path1"), test_path2=Input(path="test_path2"))
        rest_job = my_job._to_rest_object().as_dict()
        rest_items = rest_job["properties"]["jobs"]["parallelfor"]["items"]
        assert (
            rest_items == '[{"component_in_path": "${{parent.inputs.test_path1}}"}, '
            '{"component_in_path": "${{parent.inputs.test_path2}}"}]'
        )

        # dict items
        @pipeline
        def my_pipeline():
            body = basic_component(component_in_path=Input(path="test_path1"))
            parallel_for(body=body, items={"iter1": {"component_in_number": 1}, "iter2": {"component_in_number": 2}})

        my_job = my_pipeline()
        rest_job = my_job._to_rest_object().as_dict()
        rest_items = rest_job["properties"]["jobs"]["parallelfor"]["items"]
        assert rest_items == '{"iter1": {"component_in_number": 1}, "iter2": {"component_in_number": 2}}'

        # binding items
        @pipeline
        def my_pipeline(pipeline_input: str):
            body = basic_component(component_in_path=Input(path="test_path1"))
            parallel_for(body=body, items=pipeline_input)

        my_job = my_pipeline(pipeline_input='[{"component_in_number": 1}, {"component_in_number": 2}]')
        rest_job = my_job._to_rest_object().as_dict()
        rest_items = rest_job["properties"]["jobs"]["parallelfor"]["items"]
        assert rest_items == "${{parent.inputs.pipeline_input}}"

        # serialized string as items
        @pipeline
        def my_pipeline():
            body = basic_component(component_in_path=Input(path="test_path1"))
            parallel_for(body=body, items='[{"component_in_number": 1}, {"component_in_number": 2}]')

        my_job = my_pipeline()
        rest_job = my_job._to_rest_object().as_dict()
        rest_items = rest_job["properties"]["jobs"]["parallelfor"]["items"]
        assert rest_items == '[{"component_in_number": 1}, {"component_in_number": 2}]'

        # complicated input types as items
        @pipeline
        def my_pipeline():
            body = complex_component()
            parallel_for(
                body=body,
                items=[
                    dict(
                        component_in_string="component_in_string",
                        component_in_ranged_integer=10,
                        component_in_enum="world",
                        component_in_boolean=True,
                        component_in_ranged_number=5.5,
                    ),
                ],
            )

        my_job = my_pipeline()
        rest_job = my_job._to_rest_object().as_dict()
        rest_items = rest_job["properties"]["jobs"]["parallelfor"]["items"]
        assert (
            rest_items == '[{"component_in_string": "component_in_string", '
            '"component_in_ranged_integer": 10, "component_in_enum": "world", '
            '"component_in_boolean": true, "component_in_ranged_number": 5.5}]'
        )

        # JSON string items
        @pipeline
        def my_pipeline():
            body = basic_component(component_in_path=Input(path="test_path1"))
            parallel_for(body=body, items='[{"component_in_number": 1}, {"component_in_number": 2}]')

        my_job = my_pipeline()
        rest_job = my_job._to_rest_object().as_dict()
        rest_items = rest_job["properties"]["jobs"]["parallelfor"]["items"]
        assert rest_items == '[{"component_in_number": 1}, {"component_in_number": 2}]'

    @pytest.mark.parametrize(
        "output_dict, pipeline_out_dict, component_out_dict, check_pipeline_job",
        [
            ({"type": "uri_file"}, {"job_output_type": "mltable"}, {"type": "mltable"}, True),
            ({"type": "uri_folder"}, {"job_output_type": "mltable"}, {"type": "mltable"}, True),
            ({"type": "mltable"}, {"job_output_type": "mltable"}, {"type": "mltable"}, True),
            ({"type": "mlflow_model"}, {"job_output_type": "mltable"}, {"type": "mltable"}, True),
            ({"type": "triton_model"}, {"job_output_type": "mltable"}, {"type": "mltable"}, True),
            ({"type": "custom_model"}, {"job_output_type": "mltable"}, {"type": "mltable"}, True),
            ({"type": "path"}, {"job_output_type": "mltable"}, {"type": "mltable"}, True),
            ({"type": "number"}, {}, {"type": "string"}, False),
            ({"type": "string"}, {}, {"type": "string"}, False),
            ({"type": "boolean"}, {}, {"type": "string"}, False),
            ({"type": "integer"}, {}, {"type": "string"}, False),
        ],
    )
    def test_parallel_for_outputs(self, output_dict, pipeline_out_dict, component_out_dict, check_pipeline_job):
        basic_component = load_component(
            source="./tests/test_configs/components/helloworld_component.yml",
            params_override=[{"outputs.component_out_path": output_dict}],
        )

        @pipeline
        def my_pipeline():
            body = basic_component(component_in_path=Input(path="test_path1"))

            foreach_node = parallel_for(
                body=body, items={"iter1": {"component_in_number": 1}, "iter2": {"component_in_number": 2}}
            )
            return {"output": foreach_node.outputs.component_out_path}

        my_job = my_pipeline()

        if check_pipeline_job:
            rest_job = my_job._to_rest_object().as_dict()
            rest_outputs = rest_job["properties"]["outputs"]
            assert rest_outputs == {"output": pipeline_out_dict}

        pipeline_component = my_job.component
        rest_component = pipeline_component._to_rest_object().as_dict()
        assert rest_component["properties"]["component_spec"]["outputs"] == {"output": component_out_dict}

    def test_parallel_for_source(self):
        basic_component = load_component(
            source="./tests/test_configs/components/helloworld_component.yml",
        )

        @pipeline
        def my_pipeline():
            body = basic_component(component_in_path=Input(path="test_path1"))

            foreach_node = parallel_for(
                body=body, items={"iter1": {"component_in_number": 1}, "iter2": {"component_in_number": 2}}
            )

        my_job = my_pipeline()
        assert my_job.jobs["foreach_node"]._source == ComponentSource.DSL

    @pytest.mark.parametrize(
        "items, rest_input_str",
        [
            (
                # asset input with path on datastore
                {
                    "silo1": {"uri_file_input": Input(path="path/on/datastore")},
                },
                '{"silo1": {"uri_file_input": {"uri": "path/on/datastore", "job_input_type": "uri_folder"}}}',
            ),
            (
                # asset input with name + version
                {
                    "silo1": {"name_version": Input(path="test-data:1")},
                },
                '{"silo1": {"name_version": {"uri": "test-data:1", "job_input_type": "uri_folder"}}}',
            ),
            (
                # asset input with uri
                {
                    "silo1": {"uri": Input(path="https://dprepdata.blob.core.windows.net/demo/Titanic.csv")},
                },
                '{"silo1": {"uri": {"uri": "https://dprepdata.blob.core.windows.net/demo/Titanic.csv", '
                '"job_input_type": "uri_folder"}}}',
            ),
            (
                # asset input binding
                {
                    "silo1": {"binding": PipelineInput(name="input1", owner="pipeline", meta=None)},
                },
                '{"silo1": {"binding": "${{parent.inputs.input1}}"}}',
            ),
            (
                # primitive type input
                {
                    "silo1": dict(
                        component_in_string="component_in_string",
                        component_in_ranged_integer=10,
                        component_in_enum="world",
                        component_in_boolean=True,
                        component_in_ranged_number=5.5,
                    ),
                },
                '{"silo1": {"component_in_string": "component_in_string", '
                '"component_in_ranged_integer": 10, "component_in_enum": "world", '
                '"component_in_boolean": true, "component_in_ranged_number": 5.5}}',
            ),
        ],
    )
    def test_to_rest_items(self, items, rest_input_str):
        assert ParallelFor._to_rest_items(items) == rest_input_str

    @pytest.mark.parametrize(
        "items, error_message",
        [
            (
                # unsupported item value type
                [
                    {"component_in_number": CustomizedObject()},
                ],
                "Unsupported type",
            ),
            (
                # local file input
                [{"component_in_path": Input(path="./tests/test_configs/components/helloworld_component.yml")}],
                "Local file input",
            ),
            (
                # empty path
                [{"component_in_path": Input(path=None)}],
                "Input path not provided",
            ),
            (
                # dict Input
                [{"component_in_path": {"job_input_path": "azureml://path/to/file"}}],
                "Unsupported type",
            ),
        ],
    )
    def test_to_rest_items_unsupported(self, items, error_message):
        with pytest.raises(ValidationException) as e:
            ParallelFor._to_rest_items(items)
        assert error_message in str(e.value)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=protected-access

import datetime
from typing import Dict, Optional

import isodate

from azure.ai.ml._restclient.v2023_06_01_preview.models import FixedInputData as RestFixedInputData
from azure.ai.ml._restclient.v2023_06_01_preview.models import MonitoringInputDataBase as RestMonitorInputBase
from azure.ai.ml._restclient.v2023_06_01_preview.models import StaticInputData as RestStaticInputData
from azure.ai.ml._restclient.v2023_06_01_preview.models import TrailingInputData as RestTrailingInputData
from azure.ai.ml._utils.utils import camel_to_snake, snake_to_camel
from azure.ai.ml.constants._monitoring import MonitorDatasetContext, MonitorInputDataType
from azure.ai.ml.entities._mixins import RestTranslatableMixin


class MonitorInputData(RestTranslatableMixin):
    """Monitor input data.

    :keyword type: Specifies the type of monitoring input data.
    :paramtype type: MonitorInputDataType
    :keyword input_dataset: Input data used by the monitor
    :paramtype input_dataset: Optional[~azure.ai.ml.Input]
    :keyword dataset_context: The context of the input dataset. Accepted values are "model_inputs",
        "model_outputs", "training", "test", "validation", and "ground_truth".
    :paramtype dataset_context: Optional[Union[str, ~azure.ai.ml.constants.MonitorDatasetContext]]
    :keyword target_column_name: The target column in the given input dataset.
    :paramtype target_column_name: Optional[str]
    :keyword pre_processing_component: The ARM (Azure Resource Manager) resource ID of the component resource used to
        preprocess the data.
    :paramtype pre_processing_component: Optional[str]
    """

    def __init__(
        self,
        *,
        type: Optional[MonitorInputDataType] = None,
        data_context: Optional[MonitorDatasetContext] = None,
        target_columns: Optional[Dict] = None,
        job_type: Optional[str] = None,
        uri: Optional[str] = None,
    ):
        self.type = type
        self.data_context = data_context
        self.target_columns = target_columns
        self.job_type = job_type
        self.uri = uri

    @classmethod
    def _from_rest_object(cls, obj: RestMonitorInputBase) -> Optional["MonitorInputData"]:
        if obj.input_data_type == MonitorInputDataType.FIXED:
            return FixedInputData._from_rest_object(obj)
        if obj.input_data_type == MonitorInputDataType.TRAILING:
            return TrailingInputData._from_rest_object(obj)
        if obj.input_data_type == MonitorInputDataType.STATIC:
            return StaticInputData._from_rest_object(obj)

        return None


class FixedInputData(MonitorInputData):
    """
    :ivar type: Specifies the type of monitoring input data. Set automatically to "Fixed" for this class.
    :var type: MonitorInputDataType
    """

    def __init__(
        self,
        *,
        data_context: Optional[MonitorDatasetContext] = None,
        target_columns: Optional[Dict] = None,
        job_type: Optional[str] = None,
        uri: Optional[str] = None,
    ):
        super().__init__(
            type=MonitorInputDataType.FIXED,
            data_context=data_context,
            target_columns=target_columns,
            job_type=job_type,
            uri=uri,
        )

    def _to_rest_object(self) -> RestFixedInputData:
        return RestFixedInputData(
            data_context=camel_to_snake(self.data_context),
            columns=self.target_columns,
            job_input_type=self.job_type,
            uri=self.uri,
        )

    @classmethod
    def _from_rest_object(cls, obj: RestFixedInputData) -> "FixedInputData":
        return cls(
            data_context=camel_to_snake(obj.data_context),
            target_columns=obj.columns,
            job_type=obj.job_input_type,
            uri=obj.uri,
        )


class TrailingInputData(MonitorInputData):
    """
    :ivar type: Specifies the type of monitoring input data. Set automatically to "Trailing" for this class.
    :var type: MonitorInputDataType
    """

    def __init__(
        self,
        *,
        data_context: Optional[MonitorDatasetContext] = None,
        target_columns: Optional[Dict] = None,
        job_type: Optional[str] = None,
        uri: Optional[str] = None,
        window_size: Optional[str] = None,
        window_offset: Optional[str] = None,
        pre_processing_component_id: Optional[str] = None,
    ):
        super().__init__(
            type=MonitorInputDataType.TRAILING,
            data_context=data_context,
            target_columns=target_columns,
            job_type=job_type,
            uri=uri,
        )
        self.window_size = window_size
        self.window_offset = window_offset
        self.pre_processing_component_id = pre_processing_component_id

    def _to_rest_object(self) -> RestTrailingInputData:
        return RestTrailingInputData(
            data_context=camel_to_snake(self.data_context),
            columns=self.target_columns,
            job_input_type=self.job_type,
            uri=self.uri,
            window_size=self.window_size,
            window_offset=self.window_offset,
            preprocessing_component_id=self.pre_processing_component_id,
        )

    @classmethod
    def _from_rest_object(cls, obj: RestTrailingInputData) -> "TrailingInputData":
        return cls(
            data_context=snake_to_camel(obj.data_context),
            target_columns=obj.columns,
            job_type=obj.job_input_type,
            uri=obj.uri,
            window_size=str(isodate.duration_isoformat(obj.window_size)),
            window_offset=str(isodate.duration_isoformat(obj.window_offset)),
            pre_processing_component_id=obj.preprocessing_component_id,
        )


class StaticInputData(MonitorInputData):
    """
    :ivar type: Specifies the type of monitoring input data. Set automatically to "Static" for this class.
    :var type: MonitorInputDataType
    """

    def __init__(
        self,
        *,
        data_context: Optional[MonitorDatasetContext] = None,
        target_columns: Optional[Dict] = None,
        job_type: Optional[str] = None,
        uri: Optional[str] = None,
        pre_processing_component_id: Optional[str] = None,
        window_start: Optional[str] = None,
        window_end: Optional[str] = None,
    ):
        super().__init__(
            type=MonitorInputDataType.STATIC,
            data_context=data_context,
            target_columns=target_columns,
            job_type=job_type,
            uri=uri,
        )
        self.pre_processing_component_id = pre_processing_component_id
        self.window_start = window_start
        self.window_end = window_end

    def _to_rest_object(self) -> RestStaticInputData:
        return RestStaticInputData(
            data_context=camel_to_snake(self.data_context),
            columns=self.target_columns,
            job_input_type=self.job_type,
            uri=self.uri,
            preprocessing_component_id=self.pre_processing_component_id,
            window_start=datetime.datetime.strptime(str(self.window_start), "%Y-%m-%d"),
            window_end=datetime.datetime.strptime(str(self.window_end), "%Y-%m-%d"),
        )

    @classmethod
    def _from_rest_object(cls, obj: RestStaticInputData) -> "StaticInputData":
        return cls(
            data_context=snake_to_camel(obj.data_context),
            target_columns=obj.columns,
            job_type=obj.job_input_type,
            uri=obj.uri,
            pre_processing_component_id=obj.preprocessing_component_id,
            window_start=str(datetime.datetime.strftime(obj.window_start, "%Y-%m-%d")),
            window_end=datetime.datetime.strftime(obj.window_end, "%Y-%m-%d"),
        )



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

import copy
from typing import Any, Dict, List, Optional, Tuple, Type, Union

from azure.ai.ml._restclient.v2023_04_01_preview.models import JobInput as RestJobInput
from azure.ai.ml._restclient.v2023_04_01_preview.models import JobOutput as RestJobOutput
from azure.ai.ml.constants._component import ComponentJobConstants
from azure.ai.ml.entities._inputs_outputs import GroupInput, Input, Output
from azure.ai.ml.entities._util import copy_output_setting
from azure.ai.ml.exceptions import ErrorTarget, ValidationErrorType, ValidationException

from ..._input_output_helpers import (
    from_rest_data_outputs,
    from_rest_inputs_to_dataset_literal,
    to_rest_data_outputs,
    to_rest_dataset_literal_inputs,
)
from .._pipeline_job_helpers import from_dict_to_rest_io, process_sdk_component_job_io
from .attr_dict import InputsAttrDict, OutputsAttrDict, _GroupAttrDict
from .base import NodeInput, NodeOutput, PipelineInput, PipelineOutput


class NodeIOMixin:
    """Provides ability to wrap node inputs/outputs and build data bindings
    dynamically."""

    @classmethod
    def _get_supported_inputs_types(cls) -> Optional[Any]:
        return None

    @classmethod
    def _get_supported_outputs_types(cls) -> Optional[Any]:
        return None

    @classmethod
    def _validate_io(cls, value: Any, allowed_types: Optional[tuple], *, key: Optional[str] = None) -> None:
        if allowed_types is None:
            return

        if value is None or isinstance(value, allowed_types):
            pass
        else:
            msg = "Expecting {} for input/output {}, got {} instead."
            raise ValidationException(
                message=msg.format(allowed_types, key, type(value)),
                no_personal_data_message=msg.format(allowed_types, "[key]", type(value)),
                target=ErrorTarget.PIPELINE,
                error_type=ValidationErrorType.INVALID_VALUE,
            )

    def _build_input(
        self,
        name: str,
        meta: Optional[Input],
        data: Optional[Union[dict, int, bool, float, str, Output, "PipelineInput", Input]],
    ) -> NodeInput:
        # output mode of last node should not affect input mode of next node
        if isinstance(data, NodeOutput):
            # Decoupled input and output
            # value = copy.deepcopy(value)
            data = data._deepcopy()  # pylint: disable=protected-access
            data.mode = None
        elif isinstance(data, dict):
            # Use type comparison instead of is_instance to skip _GroupAttrDict
            # when loading from yaml io will be a dict,
            # like {'job_data_path': '${{parent.inputs.pipeline_job_data_path}}'}
            # parse dict to allowed type
            data = Input(**data)

            # parameter group can be of custom type, so we don't check it here
            if meta is not None and not isinstance(meta, GroupInput):
                self._validate_io(data, self._get_supported_inputs_types(), key=name)
        return NodeInput(port_name=name, meta=meta, data=data, owner=self)

    def _build_output(self, name: str, meta: Optional[Output], data: Optional[Union[Output, str]]) -> NodeOutput:
        if isinstance(data, dict):
            data = Output(**data)

        self._validate_io(data, self._get_supported_outputs_types(), key=name)
        # For un-configured outputs, settings it to None, so we won't pass extra fields(eg: default mode)
        return NodeOutput(port_name=name, meta=meta, data=data, owner=self)

    # pylint: disable=unused-argument
    def _get_default_input_val(self, val: Any):  # type: ignore
        # use None value as data placeholder for unfilled inputs.
        # server side will fill the default value
        return None

    def _build_inputs_dict(
        self,
        inputs: Dict[str, Union[Input, str, bool, int, float]],
        *,
        input_definition_dict: Optional[dict] = None,
    ) -> InputsAttrDict:
        """Build an input attribute dict so user can get/set inputs by
        accessing attribute, eg: node1.inputs.xxx.

        :param inputs: Provided kwargs when parameterizing component func.
        :type inputs: Dict[str, Union[Input, str, bool, int, float]]
        :keyword input_definition_dict: Static input definition dict. If not provided, will build inputs without meta.
        :paramtype input_definition_dict: dict
        :return: Built dynamic input attribute dict.
        :rtype: InputsAttrDict
        """
        if input_definition_dict is not None:
            # TODO: validate inputs.keys() in input_definitions.keys()
            input_dict = {}
            for key, val in input_definition_dict.items():
                if key in inputs.keys():
                    # If input is set through component functions' kwargs, create an input object with real value.
                    data = inputs[key]
                else:
                    data = self._get_default_input_val(val)  # pylint: disable=assignment-from-none

                val = self._build_input(name=key, meta=val, data=data)
                input_dict[key] = val
        else:
            input_dict = {key: self._build_input(name=key, meta=None, data=val) for key, val in inputs.items()}
        return InputsAttrDict(input_dict)

    def _build_outputs_dict(
        self, outputs: Dict, *, output_definition_dict: Optional[dict] = None, none_data: bool = False
    ) -> OutputsAttrDict:
        """Build an output attribute dict so user can get/set outputs by
        accessing attribute, eg: node1.outputs.xxx.

        :param outputs: Provided kwargs when parameterizing component func.
        :type outputs: Dict[str, Output]
        :keyword output_definition_dict: Static output definition dict.
        :paramtype output_definition_dict: Dict
        :keyword none_data: If True, will set output data to None.
        :paramtype none_data: bool
        :return: Built dynamic output attribute dict.
        :rtype: OutputsAttrDict
        """
        if output_definition_dict is not None:
            # TODO: check if we need another way to mark a un-configured output instead of just set None.
            # Create None as data placeholder for all outputs.
            output_dict = {}
            for key, val in output_definition_dict.items():
                if key in outputs.keys():
                    # If output has given value, create an output object with real value.
                    val = self._build_output(name=key, meta=val, data=outputs[key])
                else:
                    val = self._build_output(name=key, meta=val, data=None)
                output_dict[key] = val
        else:
            output_dict = {}
            for key, val in outputs.items():
                output_val = self._build_output(name=key, meta=None, data=val if not none_data else None)
                output_dict[key] = output_val
        return OutputsAttrDict(output_dict)

    def _build_inputs(self) -> Dict:
        """Build inputs of this component to a dict dict which maps output to
        actual value.

        The built input dict will have same input format as other jobs, eg:
        {
           "input_data": Input(path="path/to/input/data", mode="Mount"),
           "input_value": 10,
           "learning_rate": "${{jobs.step1.inputs.learning_rate}}"
        }

        :return: The input dict
        :rtype: Dict[str, Union[Input, str, bool, int, float]]
        """
        inputs = {}
        # pylint: disable=redefined-builtin
        for name, input in self.inputs.items():  # type: ignore
            if isinstance(input, _GroupAttrDict):
                # Flatten group inputs into inputs dict
                inputs.update(input.flatten(group_parameter_name=name))
                continue
            inputs[name] = input._to_job_input()  # pylint: disable=protected-access
        return inputs

    def _build_outputs(self) -> Dict[str, Output]:
        """Build outputs of this component to a dict which maps output to
        actual value.

        The built output dict will have same output format as other jobs, eg:
        {
            "eval_output": "${{jobs.eval.outputs.eval_output}}"
        }

        :return: The output dict
        :rtype: Dict[str, Output]
        """
        outputs = {}
        for name, output in self.outputs.items():  # type: ignore
            if isinstance(output, NodeOutput):
                output = output._to_job_output()  # pylint: disable=protected-access
            outputs[name] = output
        # Remove non-configured output
        return {k: v for k, v in outputs.items() if v is not None}

    def _to_rest_inputs(self) -> Dict[str, Dict]:
        """Translate input builders to rest input dicts.

        The built dictionary's format aligns with component job's input yaml, eg:
        {
           "input_data": {"data": {"path": "path/to/input/data"},  "mode"="Mount"},
           "input_value": 10,
           "learning_rate": "${{jobs.step1.inputs.learning_rate}}"
        }

        :return: The REST inputs
        :rtype: Dict[str, Dict]
        """
        built_inputs = self._build_inputs()
        return self._input_entity_to_rest_inputs(input_entity=built_inputs)

    @classmethod
    def _input_entity_to_rest_inputs(cls, input_entity: Dict[str, Input]) -> Dict[str, Dict]:
        # Convert io entity to rest io objects
        input_bindings, dataset_literal_inputs = process_sdk_component_job_io(
            input_entity, [ComponentJobConstants.INPUT_PATTERN]
        )

        # parse input_bindings to InputLiteral(value=str(binding))
        rest_inputs = {**input_bindings, **dataset_literal_inputs}
        # Note: The function will only be called from BaseNode,
        # and job_type is used to enable dot in pipeline job input keys,
        # so pass job_type as None directly here.
        rest_inputs = to_rest_dataset_literal_inputs(rest_inputs, job_type=None)

        # convert rest io to dict
        rest_dataset_literal_inputs = {}
        for name, val in rest_inputs.items():
            rest_dataset_literal_inputs[name] = val.as_dict()
            if hasattr(val, "mode") and val.mode:
                rest_dataset_literal_inputs[name].update({"mode": val.mode.value})
        return rest_dataset_literal_inputs

    def _to_rest_outputs(self) -> Dict[str, Dict]:
        """Translate output builders to rest output dicts.

        The built dictionary's format aligns with component job's output yaml, eg:
        {"eval_output": "${{jobs.eval.outputs.eval_output}}"}

        :return: The REST outputs
        :rtype: Dict[str, Dict]
        """
        built_outputs = self._build_outputs()

        # Convert io entity to rest io objects
        output_bindings, data_outputs = process_sdk_component_job_io(
            built_outputs, [ComponentJobConstants.OUTPUT_PATTERN]
        )
        rest_data_outputs = to_rest_data_outputs(data_outputs)

        # convert rest io to dict
        # parse output_bindings to {"value": binding, "type": "literal"} since there's no mode for it
        rest_output_bindings = {}
        for key, binding in output_bindings.items():
            rest_output_bindings[key] = {"value": binding["value"], "type": "literal"}
            if "mode" in binding:
                rest_output_bindings[key].update({"mode": binding["mode"].value})
            if "name" in binding:
                rest_output_bindings[key].update({"name": binding["name"]})
            if "version" in binding:
                rest_output_bindings[key].update({"version": binding["version"]})

        def _rename_name_and_version(output_dict: Dict) -> Dict:
            # NodeOutput can only be registered with name and version, therefore we rename here
            if "asset_name" in output_dict.keys():
                output_dict["name"] = output_dict.pop("asset_name")
            if "asset_version" in output_dict.keys():
                output_dict["version"] = output_dict.pop("asset_version")
            return output_dict

        rest_data_outputs = {name: _rename_name_and_version(val.as_dict()) for name, val in rest_data_outputs.items()}
        self._update_output_types(rest_data_outputs)
        rest_data_outputs.update(rest_output_bindings)
        return rest_data_outputs

    @classmethod
    def _from_rest_inputs(cls, inputs: Dict) -> Dict[str, Union[Input, str, bool, int, float]]:
        """Load inputs from rest inputs.

        :param inputs: The REST inputs
        :type inputs: Dict[str, Union[str, dict]]
        :return: Input dict
        :rtype: Dict[str, Union[Input, str, bool, int, float]]
        """

        # JObject -> RestJobInput/RestJobOutput
        input_bindings, rest_inputs = from_dict_to_rest_io(inputs, RestJobInput, [ComponentJobConstants.INPUT_PATTERN])

        # RestJobInput/RestJobOutput -> Input/Output
        dataset_literal_inputs = from_rest_inputs_to_dataset_literal(rest_inputs)

        return {**dataset_literal_inputs, **input_bindings}

    @classmethod
    def _from_rest_outputs(cls, outputs: Dict[str, Union[str, dict]]) -> Dict:
        """Load outputs from rest outputs.

        :param outputs: The REST outputs
        :type outputs: Dict[str, Union[str, dict]]
        :return: Output dict
        :rtype: Dict[str, Output]
        """

        # JObject -> RestJobInput/RestJobOutput
        output_bindings, rest_outputs = from_dict_to_rest_io(
            outputs, RestJobOutput, [ComponentJobConstants.OUTPUT_PATTERN]
        )

        # RestJobInput/RestJobOutput -> Input/Output
        data_outputs = from_rest_data_outputs(rest_outputs)

        return {**data_outputs, **output_bindings}

    def _update_output_types(self, rest_data_outputs: dict) -> None:
        """Update output types in rest_data_outputs according to meta level output.

        :param rest_data_outputs: The REST data outputs
        :type rest_data_outputs: Dict
        """

        for name, rest_output in rest_data_outputs.items():
            original_output = self.outputs[name]  # type: ignore
            # for configured output with meta, "correct" the output type to file to avoid the uri_folder default value
            if original_output and original_output.type:
                if original_output.type in ["AnyFile", "uri_file"]:
                    rest_output["job_output_type"] = "uri_file"


def flatten_dict(
    dct: Optional[Dict],
    _type: Union[Type["_GroupAttrDict"], Type[GroupInput]],
    *,
    allow_dict_fields: Optional[List[str]] = None,
) -> Dict:
    """Flatten inputs/input_definitions dict for inputs dict build.

    :param dct: The dictionary to flatten
    :type dct: Dict
    :param _type: Either _GroupAttrDict or GroupInput (both have the method `flatten`)
    :type _type: Union[Type["_GroupAttrDict"], Type[GroupInput]]
    :keyword allow_dict_fields: A list of keys for dictionary values that will be included in flattened output
    :paramtype allow_dict_fields: Optional[List[str]]
    :return: The flattened dict
    :rtype: Dict
    """
    _result = {}
    if dct is not None:
        for key, val in dct.items():
            # to support passing dict value as parameter group
            if allow_dict_fields and key in allow_dict_fields and isinstance(val, dict):
                # for child dict, all values are allowed to be dict
                for flattened_key, flattened_val in flatten_dict(
                    val, _type, allow_dict_fields=list(val.keys())
                ).items():
                    _result[key + "." + flattened_key] = flattened_val
                continue
            val = GroupInput.custom_class_value_to_attr_dict(val)
            if isinstance(val, _type):
                _result.update(val.flatten(group_parameter_name=key))
                continue
            _result[key] = val
    return _result


class NodeWithGroupInputMixin(NodeIOMixin):
    """This class provide build_inputs_dict for a node to use ParameterGroup as an input."""

    @classmethod
    def _validate_group_input_type(
        cls,
        input_definition_dict: dict,
        inputs: Dict[str, Union[Input, str, bool, int, float]],
    ) -> None:
        """Raise error when group input receive a value not group type.

        :param input_definition_dict: The input definition dict
        :type input_definition_dict: dict
        :param inputs: The inputs
        :type inputs: Dict[str, Union[Input, str, bool, int, float]]
        """
        # Note: We put and extra validation here instead of doing it in pipeline._validate()
        # due to group input will be discarded silently if assign it to a non-group parameter.
        group_msg = "'%s' is defined as a parameter group but got input '%s' with type '%s'."
        non_group_msg = "'%s' is defined as a parameter but got a parameter group as input."
        for key, val in inputs.items():
            definition = input_definition_dict.get(key)
            val = GroupInput.custom_class_value_to_attr_dict(val)
            if val is None:
                continue
            # 1. inputs.group = 'a string'
            if isinstance(definition, GroupInput) and not isinstance(val, (_GroupAttrDict, dict)):
                raise ValidationException(
                    message=group_msg % (key, val, type(val)),
                    no_personal_data_message=group_msg % ("[key]", "[val]", "[type(val)]"),
                    target=ErrorTarget.PIPELINE,
                    type=ValidationErrorType.INVALID_VALUE,
                )
            # 2. inputs.str_param = group
            if not isinstance(definition, GroupInput) and isinstance(val, _GroupAttrDict):
                raise ValidationException(
                    message=non_group_msg % key,
                    no_personal_data_message=non_group_msg % "[key]",
                    target=ErrorTarget.PIPELINE,
                    type=ValidationErrorType.INVALID_VALUE,
                )

    @classmethod
    def _flatten_inputs_and_definition(
        cls,
        inputs: Dict[str, Union[Input, str, bool, int, float]],
        input_definition_dict: dict,
    ) -> Tuple[Dict, Dict]:
        """
        Flatten all GroupInput(definition) and GroupAttrDict recursively and build input dict.
        For example:
        input_definition_dict = {
            "group1": GroupInput(
                values={
                    "param1": GroupInput(
                         values={
                             "param1_1": Input(type="str"),
                         }
                    ),
                    "param2": Input(type="int"),
                }
            ),
            "group2": GroupInput(
                values={
                    "param3": Input(type="str"),
                }
            ),
        } => {
            "group1.param1.param1_1": Input(type="str"),
            "group1.param2": Input(type="int"),
            "group2.param3": Input(type="str"),
        }
        inputs = {
            "group1": {
                "param1": {
                    "param1_1": "value1",
                },
                "param2": 2,
            },
            "group2": {
                "param3": "value3",
            },
        } => {
            "group1.param1.param1_1": "value1",
            "group1.param2": 2,
            "group2.param3": "value3",
        }
        :param inputs: The inputs
        :type inputs: Dict[str, Union[Input, str, bool, int, float]]
        :param input_definition_dict: The input definition dict
        :type input_definition_dict: dict
        :return: The flattened inputs and definition
        :rtype: Tuple[Dict, Dict]
        """
        group_input_names = [key for key, val in input_definition_dict.items() if isinstance(val, GroupInput)]
        flattened_inputs = flatten_dict(inputs, _GroupAttrDict, allow_dict_fields=group_input_names)
        flattened_definition_dict = flatten_dict(input_definition_dict, GroupInput)
        return flattened_inputs, flattened_definition_dict

    def _build_inputs_dict(
        self,
        inputs: Dict[str, Union[Input, str, bool, int, float]],
        *,
        input_definition_dict: Optional[dict] = None,
    ) -> InputsAttrDict:
        """Build an input attribute dict so user can get/set inputs by
        accessing attribute, eg: node1.inputs.xxx.

        :param inputs: Provided kwargs when parameterizing component func.
        :type inputs: Dict[str, Union[Input, str, bool, int, float]]
        :keyword input_definition_dict: Input definition dict from component entity.
        :paramtype input_definition_dict: dict
        :return: Built input attribute dict.
        :rtype: InputsAttrDict
        """

        # TODO: should we support group input when there is no local input definition?
        if input_definition_dict is not None:
            # Validate group mismatch
            self._validate_group_input_type(input_definition_dict, inputs)

            # Flatten inputs and definition
            flattened_inputs, flattened_definition_dict = self._flatten_inputs_and_definition(
                inputs, input_definition_dict
            )
            # Build: zip all flattened parameter with definition
            inputs = super()._build_inputs_dict(flattened_inputs, input_definition_dict=flattened_definition_dict)
            return InputsAttrDict(GroupInput.restore_flattened_inputs(inputs))
        return super()._build_inputs_dict(inputs)


class PipelineJobIOMixin(NodeWithGroupInputMixin):
    """Provides ability to wrap pipeline job inputs/outputs and build data bindings
    dynamically."""

    def _build_input(self, name: str, meta: Optional[Input], data: Any) -> "PipelineInput":
        return PipelineInput(name=name, meta=meta, data=data, owner=self)

    def _build_output(
        self, name: str, meta: Optional[Union[Input, Output]], data: Optional[Union[Output, str]]
    ) -> "PipelineOutput":
        # TODO: settings data to None for un-configured outputs so we won't passing extra fields(eg: default mode)
        result = PipelineOutput(port_name=name, meta=meta, data=data, owner=self)
        return result

    def _build_inputs_dict(
        self,
        inputs: Dict[str, Union[Input, str, bool, int, float]],
        *,
        input_definition_dict: Optional[dict] = None,
    ) -> InputsAttrDict:
        """Build an input attribute dict so user can get/set inputs by
        accessing attribute, eg: node1.inputs.xxx.

        :param inputs: Provided kwargs when parameterizing component func.
        :type inputs: Dict[str, Union[Input, str, bool, int, float]]
        :keyword input_definition_dict: Input definition dict from component entity.
        :return: Built input attribute dict.
        :rtype: InputsAttrDict
        """
        input_dict = super()._build_inputs_dict(inputs, input_definition_dict=input_definition_dict)
        # TODO: should we do this when input_definition_dict is not None?
        # TODO: should we put this in super()._build_inputs_dict?
        if input_definition_dict is None:
            return InputsAttrDict(GroupInput.restore_flattened_inputs(input_dict))
        return input_dict

    def _build_output_for_pipeline(self, name: str, data: Optional[Union[Output, NodeOutput]]) -> "PipelineOutput":
        """Build an output object for pipeline and copy settings from source output.

        :param name: Output name.
        :type name: str
        :param data: Output data.
        :type data: Optional[Union[Output, NodeOutput]]
        :return: Built output object.
        :rtype: PipelineOutput
        """
        # pylint: disable=protected-access
        if data is None:
            # For None output, build an empty output builder
            output_val = self._build_output(name=name, meta=None, data=None)
        elif isinstance(data, Output):
            # For output entity, build an output builder with data points to it
            output_val = self._build_output(name=name, meta=data, data=data)
        elif isinstance(data, NodeOutput):
            # For output builder, build a new output builder and copy settings from it
            output_val = self._build_output(name=name, meta=data._meta, data=None)
            copy_output_setting(source=data, target=output_val)
        else:
            message = "Unsupported output type: {} for pipeline output: {}: {}"
            raise ValidationException(
                message=message.format(type(data), name, data),
                no_personal_data_message=message,
                target=ErrorTarget.PIPELINE,
            )
        return output_val

    def _build_pipeline_outputs_dict(self, outputs: Dict) -> OutputsAttrDict:
        """Build an output attribute dict without output definition metadata.
        For pipeline outputs, its setting should be copied from node level outputs.

        :param outputs: Node output dict or pipeline component's outputs.
        :type outputs: Dict[str, Union[Output, NodeOutput]]
        :return: Built dynamic output attribute dict.
        :rtype: OutputsAttrDict
        """
        output_dict = {}
        for key, val in outputs.items():
            output_dict[key] = self._build_output_for_pipeline(name=key, data=val)
        return OutputsAttrDict(output_dict)

    def _build_outputs(self) -> Dict[str, Output]:
        """Build outputs of this pipeline to a dict which maps output to actual
        value.

        The built dictionary's format aligns with component job's output yaml,
        un-configured outputs will be None, eg:
        {"eval_output": "${{jobs.eval.outputs.eval_output}}", "un_configured": None}

        :return: The output dict
        :rtype: Dict[str, Output]
        """
        outputs = {}
        for name, output in self.outputs.items():  # type: ignore
            if isinstance(output, NodeOutput):
                output = output._to_job_output()  # pylint: disable=protected-access
            outputs[name] = output
        return outputs

    def _get_default_input_val(self, val: Any):  # type: ignore
        # use Default value as data placeholder for unfilled inputs.
        # client side need to fill the default value for dsl.pipeline
        if isinstance(val, GroupInput):
            # Copy default value dict for group
            return copy.deepcopy(val.default)
        return val.default

    def _update_output_types(self, rest_data_outputs: Dict) -> None:
        """Won't clear output type for pipeline level outputs since it's required in rest object.

        :param rest_data_outputs: The REST data outputs
        :type rest_data_outputs: Dict
        """


class AutoMLNodeIOMixin(NodeIOMixin):
    """Wrap outputs of automl node and build data bindings dynamically."""

    def __init__(self, **kwargs):  # type: ignore
        # add a inputs field to align with other nodes
        self.inputs = {}
        super(AutoMLNodeIOMixin, self).__init__(**kwargs)
        if getattr(self, "outputs", None):
            self._outputs = self._build_outputs_dict(self.outputs or {})



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=unused-argument

import logging
from typing import Any

from marshmallow import post_load

from azure.ai.ml._schema.core.fields import NestedField
from azure.ai.ml._schema.core.schema import PatchedSchemaMeta

from .resource_settings_schema import ResourceSettingsSchema

module_logger = logging.getLogger(__name__)


class ResourceRequirementsSchema(metaclass=PatchedSchemaMeta):
    requests = NestedField(ResourceSettingsSchema)
    limits = NestedField(ResourceSettingsSchema)

    @post_load
    def make(self, data: Any, **kwargs: Any) -> "ResourceRequirementsSettings":
        from azure.ai.ml.entities import ResourceRequirementsSettings

        return ResourceRequirementsSettings(**data)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=protected-access,arguments-renamed

import logging
from abc import abstractmethod
from os import PathLike
from typing import IO, TYPE_CHECKING, Any, AnyStr, Dict, Optional, Union

from azure.ai.ml._restclient.v2022_02_01_preview.models import BatchDeploymentData
from azure.ai.ml._restclient.v2022_05_01.models import OnlineDeploymentData
from azure.ai.ml._utils.utils import dump_yaml_to_file
from azure.ai.ml.entities._mixins import RestTranslatableMixin
from azure.ai.ml.entities._resource import Resource
from azure.ai.ml.exceptions import (
    DeploymentException,
    ErrorCategory,
    ErrorTarget,
    ValidationErrorType,
    ValidationException,
)

from .code_configuration import CodeConfiguration

# avoid circular import error
if TYPE_CHECKING:
    from azure.ai.ml.entities._assets._artifacts.model import Model
    from azure.ai.ml.entities._assets.environment import Environment

module_logger = logging.getLogger(__name__)


class Deployment(Resource, RestTranslatableMixin):
    """Endpoint Deployment base class.

    :param name: Name of the deployment resource, defaults to None
    :type name: typing.Optional[str]
    :param endpoint_name: Name of the Endpoint resource, defaults to None
    :type endpoint_name: typing.Optional[str]
    :param description: Description of the deployment resource, defaults to None
    :type description: typing.Optional[str]
    :param tags: Tag dictionary. Tags can be added, removed, and updated, defaults to None
    :type tags: typing.Optional[typing.Dict[str, typing.Any]]
    :param properties: The asset property dictionary, defaults to None
    :type properties: typing.Optional[typing.Dict[str, typing.Any]]
    :param model: The Model entity, defaults to None
    :type model: typing.Optional[typing.Union[str, ~azure.ai.ml.entities.Model]]
    :param code_configuration: Code Configuration, defaults to None
    :type code_configuration: typing.Optional[CodeConfiguration]
    :param environment: The Environment entity, defaults to None
    :type environment: typing.Optional[typing.Union[str, ~azure.ai.ml.entities.Environment]]
    :param environment_variables: Environment variables that will be set in deployment, defaults to None
    :type environment_variables: typing.Optional[typing.Dict[str, str]]
    :param code_path: Folder path to local code assets. Equivalent to code_configuration.code.path
        , defaults to None
    :type code_path: typing.Optional[typing.Union[str, PathLike]]
    :param scoring_script: Scoring script name. Equivalent to code_configuration.code.scoring_script
        , defaults to None
    :type scoring_script: typing.Optional[typing.Union[str, PathLike]]
    :raises ~azure.ai.ml.exceptions.ValidationException: Raised if Deployment cannot be successfully validated.
        Exception details will be provided in the error message.
    """

    def __init__(
        self,
        name: Optional[str] = None,
        *,
        endpoint_name: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[Dict[str, Any]] = None,
        properties: Optional[Dict[str, Any]] = None,
        model: Optional[Union[str, "Model"]] = None,
        code_configuration: Optional[CodeConfiguration] = None,
        environment: Optional[Union[str, "Environment"]] = None,
        environment_variables: Optional[Dict[str, str]] = None,
        code_path: Optional[Union[str, PathLike]] = None,
        scoring_script: Optional[Union[str, PathLike]] = None,
        **kwargs: Any,
    ):
        # MFE is case-insensitive for Name. So convert the name into lower case here.
        name = name.lower() if name else None
        self.endpoint_name = endpoint_name
        self._type: Optional[str] = kwargs.pop("type", None)

        if code_configuration and (code_path or scoring_script):
            msg = "code_path and scoring_script are not allowed if code_configuration is provided."
            raise ValidationException(
                message=msg,
                target=ErrorTarget.DEPLOYMENT,
                no_personal_data_message=msg,
                error_category=ErrorCategory.USER_ERROR,
                error_type=ValidationErrorType.INVALID_VALUE,
            )

        super().__init__(name, description, tags, properties, **kwargs)

        self.model = model
        self.code_configuration = code_configuration
        if not self.code_configuration and (code_path or scoring_script):
            self.code_configuration = CodeConfiguration(code=code_path, scoring_script=scoring_script)

        self.environment = environment
        self.environment_variables = dict(environment_variables) if environment_variables else {}

    @property
    def type(self) -> Optional[str]:
        """
        Type of deployment.

        :rtype: str
        """
        return self._type

    @property
    def code_path(self) -> Optional[Union[str, PathLike]]:
        """
        The code directory containing the scoring script.

        :rtype: Union[str, PathLike]
        """
        return self.code_configuration.code if self.code_configuration and self.code_configuration.code else None

    @code_path.setter
    def code_path(self, value: Union[str, PathLike]) -> None:
        if not self.code_configuration:
            self.code_configuration = CodeConfiguration()

        self.code_configuration.code = value

    @property
    def scoring_script(self) -> Optional[Union[str, PathLike]]:
        """
        The scoring script file path relative to the code directory.

        :rtype: Union[str, PathLike]
        """
        return self.code_configuration.scoring_script if self.code_configuration else None

    @scoring_script.setter
    def scoring_script(self, value: Union[str, PathLike]) -> None:
        if not self.code_configuration:
            self.code_configuration = CodeConfiguration()

        self.code_configuration.scoring_script = value  # type: ignore[misc]

    def dump(self, dest: Union[str, PathLike, IO[AnyStr]], **kwargs: Any) -> None:
        """Dump the deployment content into a file in yaml format.

        :param dest: The destination to receive this deployment's content.
            Must be either a path to a local file, or an already-open file stream.
            If dest is a file path, a new file will be created,
            and an exception is raised if the file exists.
            If dest is an open file, the file will be written to directly,
            and an exception will be raised if the file is not writable.
        :type dest: typing.Union[os.PathLike, str, typing.IO[typing.AnyStr]]
        """
        path = kwargs.pop("path", None)
        yaml_serialized = self._to_dict()
        dump_yaml_to_file(dest, yaml_serialized, default_flow_style=False, path=path, **kwargs)

    @abstractmethod
    def _to_dict(self) -> Dict:
        pass

    @classmethod
    def _from_rest_object(
        cls, deployment_rest_object: Union[OnlineDeploymentData, BatchDeploymentData]
    ) -> Union[OnlineDeploymentData, BatchDeploymentData]:
        from azure.ai.ml.entities._deployment.batch_deployment import BatchDeployment
        from azure.ai.ml.entities._deployment.online_deployment import OnlineDeployment

        if isinstance(deployment_rest_object, OnlineDeploymentData):
            return OnlineDeployment._from_rest_object(deployment_rest_object)
        if isinstance(deployment_rest_object, BatchDeploymentData):
            return BatchDeployment._from_rest_object(deployment_rest_object)

        msg = f"Unsupported deployment type {type(deployment_rest_object)}"
        raise DeploymentException(
            message=msg,
            target=ErrorTarget.DEPLOYMENT,
            no_personal_data_message=msg,
            error_category=ErrorCategory.SYSTEM_ERROR,
        )

    def _to_rest_object(self) -> Any:
        pass

    def _merge_with(self, other: "Deployment") -> None:
        if other:
            if self.name != other.name:
                msg = "The deployment name: {} and {} are not matched when merging."
                raise ValidationException(
                    message=msg.format(self.name, other.name),
                    target=ErrorTarget.DEPLOYMENT,
                    no_personal_data_message=msg.format("[name1]", "[name2]"),
                    error_category=ErrorCategory.USER_ERROR,
                    error_type=ValidationErrorType.INVALID_VALUE,
                )
            if other.tags:
                self.tags: dict = {**self.tags, **other.tags}
            if other.properties:
                self.properties: dict = {**self.properties, **other.properties}
            if other.environment_variables:
                self.environment_variables = {
                    **self.environment_variables,
                    **other.environment_variables,
                }
            self.code_configuration = other.code_configuration or self.code_configuration
            self.model = other.model or self.model
            self.environment = other.environment or self.environment
            self.endpoint_name = other.endpoint_name or self.endpoint_name



from os import environ

import pytest
from marshmallow.exceptions import ValidationError

from azure.ai.ml import load_component, load_job
from azure.ai.ml.constants import ImportSourceType
from azure.ai.ml.constants._common import AZUREML_PRIVATE_FEATURES_ENV_VAR, AssetTypes
from azure.ai.ml.entities._builders.import_node import Import
from azure.ai.ml.entities._component.import_component import ImportComponent
from azure.ai.ml.entities._inputs_outputs import Output
from azure.ai.ml.entities._job.import_job import DatabaseImportSource, FileImportSource, ImportJob, ImportSource


@pytest.mark.unittest
class TestImportJob:
    environ[AZUREML_PRIVATE_FEATURES_ENV_VAR] = "True"

    @pytest.mark.parametrize(
        "yaml_file,expected_import_source,expected_import_source_type",
        [
            ("import_job_test.yml", DatabaseImportSource, ImportSourceType.AZURESQLDB),
            ("import_job_test_azuresynapseanalytics.yml", DatabaseImportSource, ImportSourceType.AZURESYNAPSEANALYTICS),
            ("import_job_test_snowflake.yml", DatabaseImportSource, ImportSourceType.SNOWFLAKE),
            ("import_job_test_s3.yml", FileImportSource, ImportSourceType.S3),
        ],
    )
    def test_import_job(self, yaml_file: str, expected_import_source: ImportSource, expected_import_source_type: str):
        import_job = load_job(
            source="./tests/test_configs/import_job/" + yaml_file,
        )

        source = (
            DatabaseImportSource(
                type=expected_import_source_type,
                connection="azureml:my_username_password",
                query="select * from REGION",
            )
            if expected_import_source is DatabaseImportSource
            else FileImportSource(
                type=expected_import_source_type, connection="azureml:my_username_password", path="test1/*"
            )
        )
        output_type = AssetTypes.MLTABLE if expected_import_source is DatabaseImportSource else AssetTypes.URI_FOLDER

        import_job2 = ImportJob(
            description="test_description",
            display_name="test_display_name",
            source=source,
            output=Output(type=output_type, path="azureml://datastores/workspaceblobstore/paths/output_dir/"),
        )

        assert import_job.description == import_job2.description
        assert import_job.display_name == import_job2.display_name

        assert import_job.output.type == import_job2.output.type
        assert import_job.output.path == import_job2.output.path

        assert isinstance(import_job, ImportJob)
        assert isinstance(import_job.source, expected_import_source)
        assert import_job.source.type == import_job2.source.type
        assert import_job.source.connection == import_job2.source.connection

    @pytest.mark.parametrize(
        "yaml_file,expected_exception_messages",
        [
            (
                "import_job_test_missing_keys.yml",
                ["source: [Missing data for required field.]", "output: [Missing data for required field.]"],
            ),
            (
                "import_job_test_missing_fields.yml",
                ["connection: [Missing data for required field.]", "query: [Missing data for required field.]"],
            ),
            (
                "import_job_test_unknown_keys.yml",
                ["compute: [Unknown field.]", "inputs: [Unknown field.]", "outputs: [Unknown field.]"],
            ),
            (
                "import_job_test_unknown_fields.yml",
                ["query: [Missing data for required field.]", "path: [Unknown field.]"],
            ),
        ],
    )
    def test_import_job_missing_unknown_fields(self, yaml_file: str, expected_exception_messages: list):
        with pytest.raises(ValidationError) as exception_raised:
            load_job("./tests/test_configs/import_job/" + yaml_file)

        assert exception_raised.value.messages is not None and len(exception_raised.value.messages) > 0
        assert "Validation for ImportJobSchema failed" in exception_raised.value.messages[0]

        exception_message = exception_raised.value.messages[0].replace("\n", "").replace('"', "").replace("  ", "")

        for expected_exception_message in expected_exception_messages:
            assert expected_exception_message in exception_message

    def test_import_component(self):
        import_func = load_component("./tests/test_configs/import_job/import_component_test.yml")
        import_node: Import = import_func(
            type="azuresqldb", connection="azureml:my_username_password", query="select * from REGION"
        )
        import_job: ImportJob = import_node._to_job()

        assert isinstance(import_node, Import)
        assert isinstance(import_node.component, ImportComponent)
        assert isinstance(import_job.source, DatabaseImportSource)

        assert import_node.component.description == "This is the basic import component"
        assert import_node.component.display_name == "ImportComponentBasic"

        assert import_job.source.type == "azuresqldb"
        assert import_job.source.connection == "azureml:my_username_password"
        assert import_job.source.query == "select * from REGION"

        assert import_node.inputs.get("type")._data == import_job.source.type
        assert import_node.inputs.get("connection")._data == import_job.source.connection
        assert import_node.inputs.get("query")._data == import_job.source.query
        assert import_node.outputs.get("output").type == AssetTypes.MLTABLE

    @pytest.mark.parametrize(
        "yaml_file,expected_exception_messages",
        [
            (
                "import_component_test_missing_keys.yml",
                {"source": "Missing data for required field.", "output": "Missing data for required field."},
            ),
        ],
    )
    def test_import_component_missing_unknown_fields(self, yaml_file: str, expected_exception_messages: dict):
        with pytest.raises(ValidationError) as exception_raised:
            load_component(source="./tests/test_configs/import_job/" + yaml_file)

        assert exception_raised.value.messages is not None and len(exception_raised.value.messages) > 0
        assert exception_raised.typename == "ValidationError"

        for key in expected_exception_messages:
            assert any(key in msg for msg in exception_raised.value.messages)
            # assert expected_exception_messages[key] in exception_raised.value.messages[key]



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
from pathlib import Path
from typing import Optional

from ...constants._common import AzureDevopsArtifactsType
from ...entities._component._additional_includes import AdditionalIncludes


class InternalAdditionalIncludes(AdditionalIncludes):
    """This class is kept for compatibility with mldesigner."""

    @property
    def includes(self):
        if self._is_artifact_includes:
            return self._get_resolved_additional_include_configs()
        return self.origin_configs

    @property
    def code_path(self) -> Optional[Path]:
        return self.resolved_code_path

    @property
    def _is_artifact_includes(self):
        return any(
            map(
                lambda x: isinstance(x, dict) and x.get("type", None) == AzureDevopsArtifactsType.ARTIFACT,
                self.origin_configs,
            )
        )



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

__path__ = __import__("pkgutil").extend_path(__path__, __name__)

from .distillation_job import DistillationJobSchema
from .endpoint_request_settings import EndpointRequestSettingsSchema
from .prompt_settings import PromptSettingsSchema
from .teacher_model_settings import TeacherModelSettingsSchema

__all__ = [
    "DistillationJobSchema",
    "PromptSettingsSchema",
    "EndpointRequestSettingsSchema",
    "TeacherModelSettingsSchema",
]



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
# pylint: disable=protected-access
from datetime import datetime, timezone
from typing import Any, Iterable, List, Optional, Tuple, cast

from azure.ai.ml._restclient.v2023_06_01_preview import AzureMachineLearningWorkspaces as ServiceClient062023Preview
from azure.ai.ml._restclient.v2024_01_01_preview import AzureMachineLearningWorkspaces as ServiceClient012024Preview
from azure.ai.ml._scope_dependent_operations import (
    OperationConfig,
    OperationsContainer,
    OperationScope,
    _ScopeDependentOperations,
)
from azure.ai.ml._telemetry import ActivityType, monitor_with_activity, monitor_with_telemetry_mixin
from azure.ai.ml._utils._logger_utils import OpsLogger
from azure.ai.ml.entities import Job, JobSchedule, Schedule
from azure.ai.ml.entities._inputs_outputs.input import Input
from azure.ai.ml.entities._monitoring.schedule import MonitorSchedule
from azure.ai.ml.entities._monitoring.signals import (
    BaselineDataRange,
    FADProductionData,
    GenerationTokenStatisticsSignal,
    LlmData,
    ProductionData,
    ReferenceData,
)
from azure.ai.ml.entities._monitoring.target import MonitoringTarget
from azure.ai.ml.exceptions import ErrorCategory, ErrorTarget, ScheduleException
from azure.core.credentials import TokenCredential
from azure.core.polling import LROPoller
from azure.core.tracing.decorator import distributed_trace

from .._restclient.v2022_10_01.models import ScheduleListViewType
from .._restclient.v2024_01_01_preview.models import TriggerOnceRequest
from .._utils._arm_id_utils import AMLNamedArmId, AMLVersionedArmId, is_ARM_id_for_parented_resource
from .._utils._azureml_polling import AzureMLPolling
from .._utils.utils import snake_to_camel
from ..constants._common import (
    ARM_ID_PREFIX,
    AZUREML_RESOURCE_PROVIDER,
    NAMED_RESOURCE_ID_FORMAT_WITH_PARENT,
    AzureMLResourceType,
    LROConfigurations,
)
from ..constants._monitoring import (
    DEPLOYMENT_MODEL_INPUTS_COLLECTION_KEY,
    DEPLOYMENT_MODEL_INPUTS_NAME_KEY,
    DEPLOYMENT_MODEL_INPUTS_VERSION_KEY,
    DEPLOYMENT_MODEL_OUTPUTS_COLLECTION_KEY,
    DEPLOYMENT_MODEL_OUTPUTS_NAME_KEY,
    DEPLOYMENT_MODEL_OUTPUTS_VERSION_KEY,
    MonitorDatasetContext,
    MonitorSignalType,
)
from ..entities._schedule.schedule import ScheduleTriggerResult
from . import DataOperations, JobOperations, OnlineDeploymentOperations
from ._job_ops_helper import stream_logs_until_completion
from ._operation_orchestrator import OperationOrchestrator

ops_logger = OpsLogger(__name__)
module_logger = ops_logger.module_logger


class ScheduleOperations(_ScopeDependentOperations):
    # pylint: disable=too-many-instance-attributes
    """
    ScheduleOperations

    You should not instantiate this class directly.
    Instead, you should create an MLClient instance that instantiates it for you and attaches it as an attribute.
    """

    def __init__(
        self,
        operation_scope: OperationScope,
        operation_config: OperationConfig,
        service_client_06_2023_preview: ServiceClient062023Preview,
        service_client_01_2024_preview: ServiceClient012024Preview,
        all_operations: OperationsContainer,
        credential: TokenCredential,
        **kwargs: Any,
    ):
        super(ScheduleOperations, self).__init__(operation_scope, operation_config)
        ops_logger.update_filter()
        self.service_client = service_client_06_2023_preview.schedules
        # Note: Trigger once is supported since 24_01, we don't upgrade other operations' client because there are
        # some breaking changes, for example: AzMonMonitoringAlertNotificationSettings is removed.
        self.schedule_trigger_service_client = service_client_01_2024_preview.schedules
        self._all_operations = all_operations
        self._stream_logs_until_completion = stream_logs_until_completion
        # Dataplane service clients are lazily created as they are needed
        self._runs_operations_client = None
        self._dataset_dataplane_operations_client = None
        self._model_dataplane_operations_client = None
        # Kwargs to propagate to dataplane service clients
        self._service_client_kwargs = kwargs.pop("_service_client_kwargs", {})
        self._api_base_url = None
        self._container = "azureml"
        self._credential = credential
        self._orchestrators = OperationOrchestrator(self._all_operations, self._operation_scope, self._operation_config)

        self._kwargs = kwargs

    @property
    def _job_operations(self) -> JobOperations:
        return cast(
            JobOperations,
            self._all_operations.get_operation(  # type: ignore[misc]
                AzureMLResourceType.JOB, lambda x: isinstance(x, JobOperations)
            ),
        )

    @property
    def _online_deployment_operations(self) -> OnlineDeploymentOperations:
        return cast(
            OnlineDeploymentOperations,
            self._all_operations.get_operation(  # type: ignore[misc]
                AzureMLResourceType.ONLINE_DEPLOYMENT, lambda x: isinstance(x, OnlineDeploymentOperations)
            ),
        )

    @property
    def _data_operations(self) -> DataOperations:
        return cast(
            DataOperations,
            self._all_operations.get_operation(  # type: ignore[misc]
                AzureMLResourceType.DATA, lambda x: isinstance(x, DataOperations)
            ),
        )

    @distributed_trace
    @monitor_with_activity(ops_logger, "Schedule.List", ActivityType.PUBLICAPI)
    def list(
        self,
        *,
        list_view_type: ScheduleListViewType = ScheduleListViewType.ENABLED_ONLY,
        **kwargs: Any,
    ) -> Iterable[Schedule]:
        """List schedules in specified workspace.

        :keyword list_view_type: View type for including/excluding (for example)
            archived schedules. Default: ENABLED_ONLY.
        :type list_view_type: Optional[ScheduleListViewType]
        :return: An iterator to list Schedule.
        :rtype: Iterable[Schedule]
        """

        def safe_from_rest_object(objs: Any) -> List:
            result = []
            for obj in objs:
                try:
                    result.append(Schedule._from_rest_object(obj))
                except Exception as e:  # pylint: disable=W0718
                    print(f"Translate {obj.name} to Schedule failed with: {e}")
            return result

        return cast(
            Iterable[Schedule],
            self.service_client.list(
                resource_group_name=self._operation_scope.resource_group_name,
                workspace_name=self._workspace_name,
                list_view_type=list_view_type,
                cls=safe_from_rest_object,
                **self._kwargs,
                **kwargs,
            ),
        )

    def _get_polling(self, name: Optional[str]) -> AzureMLPolling:
        """Return the polling with custom poll interval.

        :param name: The schedule name
        :type name: str
        :return: The AzureMLPolling object
        :rtype: AzureMLPolling
        """
        path_format_arguments = {
            "scheduleName": name,
            "resourceGroupName": self._resource_group_name,
            "workspaceName": self._workspace_name,
        }
        return AzureMLPolling(
            LROConfigurations.POLL_INTERVAL,
            path_format_arguments=path_format_arguments,
        )

    @distributed_trace
    @monitor_with_activity(ops_logger, "Schedule.Delete", ActivityType.PUBLICAPI)
    def begin_delete(
        self,
        name: str,
        **kwargs: Any,
    ) -> LROPoller[None]:
        """Delete schedule.

        :param name: Schedule name.
        :type name: str
        :return: A poller for deletion status
        :rtype: LROPoller[None]
        """
        poller = self.service_client.begin_delete(
            resource_group_name=self._operation_scope.resource_group_name,
            workspace_name=self._workspace_name,
            name=name,
            polling=self._get_polling(name),
            **self._kwargs,
            **kwargs,
        )
        return poller

    @distributed_trace
    @monitor_with_telemetry_mixin(ops_logger, "Schedule.Get", ActivityType.PUBLICAPI)
    def get(
        self,
        name: str,
        **kwargs: Any,
    ) -> Schedule:
        """Get a schedule.

        :param name: Schedule name.
        :type name: str
        :return: The schedule object.
        :rtype: Schedule
        """
        return self.service_client.get(
            resource_group_name=self._operation_scope.resource_group_name,
            workspace_name=self._workspace_name,
            name=name,
            cls=lambda _, obj, __: Schedule._from_rest_object(obj),
            **self._kwargs,
            **kwargs,
        )

    @distributed_trace
    @monitor_with_telemetry_mixin(ops_logger, "Schedule.CreateOrUpdate", ActivityType.PUBLICAPI)
    def begin_create_or_update(
        self,
        schedule: Schedule,
        **kwargs: Any,
    ) -> LROPoller[Schedule]:
        """Create or update schedule.

        :param schedule: Schedule definition.
        :type schedule: Schedule
        :return: An instance of LROPoller that returns Schedule if no_wait=True, or Schedule if no_wait=False
        :rtype: Union[LROPoller, Schedule]
        :rtype: Union[LROPoller, ~azure.ai.ml.entities.Schedule]
        """

        if isinstance(schedule, JobSchedule):
            schedule._validate(raise_error=True)
            if isinstance(schedule.create_job, Job):
                # Create all dependent resources for job inside schedule
                self._job_operations._resolve_arm_id_or_upload_dependencies(schedule.create_job)
        elif isinstance(schedule, MonitorSchedule):
            # resolve ARM id for target, compute, and input datasets for each signal
            self._resolve_monitor_schedule_arm_id(schedule)
        # Create schedule
        schedule_data = schedule._to_rest_object()  # type: ignore
        poller = self.service_client.begin_create_or_update(
            resource_group_name=self._operation_scope.resource_group_name,
            workspace_name=self._workspace_name,
            name=schedule.name,
            cls=lambda _, obj, __: Schedule._from_rest_object(obj),
            body=schedule_data,
            polling=self._get_polling(schedule.name),
            **self._kwargs,
            **kwargs,
        )
        return poller

    @distributed_trace
    @monitor_with_activity(ops_logger, "Schedule.Enable", ActivityType.PUBLICAPI)
    def begin_enable(
        self,
        name: str,
        **kwargs: Any,
    ) -> LROPoller[Schedule]:
        """Enable a schedule.

        :param name: Schedule name.
        :type name: str
        :return: An instance of LROPoller that returns Schedule
        :rtype: LROPoller
        """
        schedule = self.get(name=name)
        schedule._is_enabled = True
        return self.begin_create_or_update(schedule, **kwargs)

    @distributed_trace
    @monitor_with_activity(ops_logger, "Schedule.Disable", ActivityType.PUBLICAPI)
    def begin_disable(
        self,
        name: str,
        **kwargs: Any,
    ) -> LROPoller[Schedule]:
        """Disable a schedule.

        :param name: Schedule name.
        :type name: str
        :return: An instance of LROPoller that returns Schedule if no_wait=True, or Schedule if no_wait=False
        :rtype:  LROPoller
        """
        schedule = self.get(name=name)
        schedule._is_enabled = False
        return self.begin_create_or_update(schedule, **kwargs)

    @distributed_trace
    @monitor_with_activity(ops_logger, "Schedule.Trigger", ActivityType.PUBLICAPI)
    def trigger(
        self,
        name: str,
        **kwargs: Any,
    ) -> ScheduleTriggerResult:
        """Trigger a schedule once.

        :param name: Schedule name.
        :type name: str
        :return: TriggerRunSubmissionDto, or the result of cls(response)
        :rtype: ~azure.ai.ml.entities.ScheduleTriggerResult
        """
        schedule_time = kwargs.pop("schedule_time", datetime.now(timezone.utc).isoformat())
        return self.schedule_trigger_service_client.trigger(
            name=name,
            resource_group_name=self._operation_scope.resource_group_name,
            workspace_name=self._workspace_name,
            body=TriggerOnceRequest(schedule_time=schedule_time),
            cls=lambda _, obj, __: ScheduleTriggerResult._from_rest_object(obj),
            **kwargs,
        )

    def _resolve_monitor_schedule_arm_id(  # pylint:disable=too-many-branches,too-many-statements,too-many-locals
        self, schedule: MonitorSchedule
    ) -> None:
        # resolve target ARM ID
        model_inputs_name, model_outputs_name = None, None
        app_traces_name, app_traces_version = None, None
        model_inputs_version, model_outputs_version = None, None
        mdc_input_enabled, mdc_output_enabled = False, False
        target = schedule.create_monitor.monitoring_target
        if target and target.endpoint_deployment_id:
            endpoint_name, deployment_name = self._process_and_get_endpoint_deployment_names_from_id(target)
            online_deployment = self._online_deployment_operations.get(deployment_name, endpoint_name)
            deployment_data_collector = online_deployment.data_collector
            if deployment_data_collector:
                in_reg = AMLVersionedArmId(deployment_data_collector.collections.get("model_inputs").data)
                out_reg = AMLVersionedArmId(deployment_data_collector.collections.get("model_outputs").data)
                if "app_traces" in deployment_data_collector.collections:
                    app_traces = AMLVersionedArmId(deployment_data_collector.collections.get("app_traces").data)
                    app_traces_name = app_traces.asset_name
                    app_traces_version = app_traces.asset_version
                model_inputs_name = in_reg.asset_name
                model_inputs_version = in_reg.asset_version
                model_outputs_name = out_reg.asset_name
                model_outputs_version = out_reg.asset_version
                mdc_input_enabled_str = deployment_data_collector.collections.get("model_inputs").enabled
                mdc_output_enabled_str = deployment_data_collector.collections.get("model_outputs").enabled
            else:
                model_inputs_name = online_deployment.tags.get(DEPLOYMENT_MODEL_INPUTS_NAME_KEY)
                model_inputs_version = online_deployment.tags.get(DEPLOYMENT_MODEL_INPUTS_VERSION_KEY)
                model_outputs_name = online_deployment.tags.get(DEPLOYMENT_MODEL_OUTPUTS_NAME_KEY)
                model_outputs_version = online_deployment.tags.get(DEPLOYMENT_MODEL_OUTPUTS_VERSION_KEY)
                mdc_input_enabled_str = online_deployment.tags.get(DEPLOYMENT_MODEL_INPUTS_COLLECTION_KEY)
                mdc_output_enabled_str = online_deployment.tags.get(DEPLOYMENT_MODEL_OUTPUTS_COLLECTION_KEY)
            if mdc_input_enabled_str and mdc_input_enabled_str.lower() == "true":
                mdc_input_enabled = True
            if mdc_output_enabled_str and mdc_output_enabled_str.lower() == "true":
                mdc_output_enabled = True
        elif target and target.model_id:
            target.model_id = self._orchestrators.get_asset_arm_id(  # type: ignore
                target.model_id,
                AzureMLResourceType.MODEL,
                register_asset=False,
            )

        if not schedule.create_monitor.monitoring_signals:
            if mdc_input_enabled and mdc_output_enabled:
                schedule._create_default_monitor_definition()
            else:
                msg = (
                    "An ARM id for a deployment with data collector enabled for model inputs and outputs must be "
                    "given if monitoring_signals is None"
                )
                raise ScheduleException(
                    message=msg,
                    no_personal_data_message=msg,
                    target=ErrorTarget.SCHEDULE,
                    error_category=ErrorCategory.USER_ERROR,
                )
        # resolve ARM id for each signal and populate any defaults if needed
        for signal_name, signal in schedule.create_monitor.monitoring_signals.items():  # type: ignore
            if signal.type == MonitorSignalType.GENERATION_SAFETY_QUALITY:
                for llm_data in signal.production_data:  # type: ignore[union-attr]
                    self._job_operations._resolve_job_input(llm_data.input_data, schedule._base_path)
                continue
            if signal.type == MonitorSignalType.GENERATION_TOKEN_STATISTICS:
                if not signal.production_data:  # type: ignore[union-attr]
                    # if target dataset is absent and data collector for input is enabled,
                    # create a default target dataset with production app traces as target
                    if isinstance(signal, GenerationTokenStatisticsSignal):
                        signal.production_data = LlmData(  # type: ignore[union-attr]
                            input_data=Input(
                                path=f"{app_traces_name}:{app_traces_version}",
                                type=self._data_operations.get(app_traces_name, app_traces_version).type,
                            ),
                            data_window=BaselineDataRange(lookback_window_size="P7D", lookback_window_offset="P0D"),
                        )
                self._job_operations._resolve_job_input(
                    signal.production_data.input_data, schedule._base_path  # type: ignore[union-attr]
                )
                continue
            if signal.type == MonitorSignalType.CUSTOM:
                if signal.inputs:  # type: ignore[union-attr]
                    for inputs in signal.inputs.values():  # type: ignore[union-attr]
                        self._job_operations._resolve_job_input(inputs, schedule._base_path)
                for data in signal.input_data.values():  # type: ignore[union-attr]
                    if data.input_data is not None:
                        for inputs in data.input_data.values():
                            self._job_operations._resolve_job_input(inputs, schedule._base_path)
                    data.pre_processing_component = self._orchestrators.get_asset_arm_id(
                        asset=data.pre_processing_component if hasattr(data, "pre_processing_component") else None,
                        azureml_type=AzureMLResourceType.COMPONENT,
                    )
                continue
            error_messages = []
            if not signal.production_data or not signal.reference_data:  # type: ignore[union-attr]
                # if there is no target dataset, we check the type of signal
                if signal.type in {MonitorSignalType.DATA_DRIFT, MonitorSignalType.DATA_QUALITY}:
                    if mdc_input_enabled:
                        if not signal.production_data:  # type: ignore[union-attr]
                            # if target dataset is absent and data collector for input is enabled,
                            # create a default target dataset with production model inputs as target
                            signal.production_data = ProductionData(  # type: ignore[union-attr]
                                input_data=Input(
                                    path=f"{model_inputs_name}:{model_inputs_version}",
                                    type=self._data_operations.get(model_inputs_name, model_inputs_version).type,
                                ),
                                data_context=MonitorDatasetContext.MODEL_INPUTS,
                                data_window=BaselineDataRange(
                                    lookback_window_size="default", lookback_window_offset="P0D"
                                ),
                            )
                        if not signal.reference_data:  # type: ignore[union-attr]
                            signal.reference_data = ReferenceData(  # type: ignore[union-attr]
                                input_data=Input(
                                    path=f"{model_inputs_name}:{model_inputs_version}",
                                    type=self._data_operations.get(model_inputs_name, model_inputs_version).type,
                                ),
                                data_context=MonitorDatasetContext.MODEL_INPUTS,
                                data_window=BaselineDataRange(
                                    lookback_window_size="default", lookback_window_offset="default"
                                ),
                            )
                    elif not mdc_input_enabled and not (
                        signal.production_data and signal.reference_data  # type: ignore[union-attr]
                    ):
                        # if target or baseline dataset is absent and data collector for input is not enabled,
                        # collect exception message
                        msg = (
                            f"A target and baseline dataset must be provided for signal with name {signal_name}"
                            f"and type {signal.type} if the monitoring_target endpoint_deployment_id is empty"
                            "or refers to a deployment for which data collection for model inputs is not enabled."
                        )
                        error_messages.append(msg)
                elif signal.type == MonitorSignalType.PREDICTION_DRIFT:
                    if mdc_output_enabled:
                        if not signal.production_data:  # type: ignore[union-attr]
                            # if target dataset is absent and data collector for output is enabled,
                            # create a default target dataset with production model outputs as target
                            signal.production_data = ProductionData(  # type: ignore[union-attr]
                                input_data=Input(
                                    path=f"{model_outputs_name}:{model_outputs_version}",
                                    type=self._data_operations.get(model_outputs_name, model_outputs_version).type,
                                ),
                                data_context=MonitorDatasetContext.MODEL_OUTPUTS,
                                data_window=BaselineDataRange(
                                    lookback_window_size="default", lookback_window_offset="P0D"
                                ),
                            )
                        if not signal.reference_data:  # type: ignore[union-attr]
                            signal.reference_data = ReferenceData(  # type: ignore[union-attr]
                                input_data=Input(
                                    path=f"{model_outputs_name}:{model_outputs_version}",
                                    type=self._data_operations.get(model_outputs_name, model_outputs_version).type,
                                ),
                                data_context=MonitorDatasetContext.MODEL_OUTPUTS,
                                data_window=BaselineDataRange(
                                    lookback_window_size="default", lookback_window_offset="default"
                                ),
                            )
                    elif not mdc_output_enabled and not (
                        signal.production_data and signal.reference_data  # type: ignore[union-attr]
                    ):
                        # if target dataset is absent and data collector for output is not enabled,
                        # collect exception message
                        msg = (
                            f"A target and baseline dataset must be provided for signal with name {signal_name}"
                            f"and type {signal.type} if the monitoring_target endpoint_deployment_id is empty"
                            "or refers to a deployment for which data collection for model outputs is not enabled."
                        )
                        error_messages.append(msg)
                elif signal.type == MonitorSignalType.FEATURE_ATTRIBUTION_DRIFT:
                    if mdc_input_enabled:
                        if not signal.production_data:  # type: ignore[union-attr]
                            # if production dataset is absent and data collector for input is enabled,
                            # create a default prod dataset with production model inputs and outputs as target
                            signal.production_data = [  # type: ignore[union-attr]
                                FADProductionData(
                                    input_data=Input(
                                        path=f"{model_inputs_name}:{model_inputs_version}",
                                        type=self._data_operations.get(model_inputs_name, model_inputs_version).type,
                                    ),
                                    data_context=MonitorDatasetContext.MODEL_INPUTS,
                                    data_window=BaselineDataRange(
                                        lookback_window_size="default", lookback_window_offset="P0D"
                                    ),
                                ),
                                FADProductionData(
                                    input_data=Input(
                                        path=f"{model_outputs_name}:{model_outputs_version}",
                                        type=self._data_operations.get(model_outputs_name, model_outputs_version).type,
                                    ),
                                    data_context=MonitorDatasetContext.MODEL_OUTPUTS,
                                    data_window=BaselineDataRange(
                                        lookback_window_size="default", lookback_window_offset="P0D"
                                    ),
                                ),
                            ]
                    elif not mdc_output_enabled and not signal.production_data:  # type: ignore[union-attr]
                        # if target dataset is absent and data collector for output is not enabled,
                        # collect exception message
                        msg = (
                            f"A production data must be provided for signal with name {signal_name}"
                            f"and type {signal.type} if the monitoring_target endpoint_deployment_id is empty"
                            "or refers to a deployment for which data collection for model outputs is not enabled."
                        )
                        error_messages.append(msg)
            if error_messages:
                # if any error messages, raise an exception with all of them so user knows which signals
                # need to be fixed
                msg = "\n".join(error_messages)
                raise ScheduleException(
                    message=msg,
                    no_personal_data_message=msg,
                    ErrorTarget=ErrorTarget.SCHEDULE,
                    ErrorCategory=ErrorCategory.USER_ERROR,
                )
            if signal.type == MonitorSignalType.FEATURE_ATTRIBUTION_DRIFT:
                for prod_data in signal.production_data:  # type: ignore[union-attr]
                    self._job_operations._resolve_job_input(prod_data.input_data, schedule._base_path)
                    prod_data.pre_processing_component = self._orchestrators.get_asset_arm_id(  # type: ignore
                        asset=prod_data.pre_processing_component,  # type: ignore[union-attr]
                        azureml_type=AzureMLResourceType.COMPONENT,
                    )
                self._job_operations._resolve_job_input(
                    signal.reference_data.input_data, schedule._base_path  # type: ignore[union-attr]
                )
                signal.reference_data.pre_processing_component = self._orchestrators.get_asset_arm_id(  # type: ignore
                    asset=signal.reference_data.pre_processing_component,  # type: ignore[union-attr]
                    azureml_type=AzureMLResourceType.COMPONENT,
                )
                continue

            self._job_operations._resolve_job_inputs(
                [signal.production_data.input_data, signal.reference_data.input_data],  # type: ignore[union-attr]
                schedule._base_path,
            )
            signal.production_data.pre_processing_component = self._orchestrators.get_asset_arm_id(  # type: ignore
                asset=signal.production_data.pre_processing_component,  # type: ignore[union-attr]
                azureml_type=AzureMLResourceType.COMPONENT,
            )
            signal.reference_data.pre_processing_component = self._orchestrators.get_asset_arm_id(  # type: ignore
                asset=signal.reference_data.pre_processing_component,  # type: ignore[union-attr]
                azureml_type=AzureMLResourceType.COMPONENT,
            )

    def _process_and_get_endpoint_deployment_names_from_id(self, target: MonitoringTarget) -> Tuple:
        target.endpoint_deployment_id = (
            target.endpoint_deployment_id[len(ARM_ID_PREFIX) :]  # type: ignore
            if target.endpoint_deployment_id is not None and target.endpoint_deployment_id.startswith(ARM_ID_PREFIX)
            else target.endpoint_deployment_id
        )

        # if it is an ARM ID, don't process it
        if not is_ARM_id_for_parented_resource(
            target.endpoint_deployment_id,
            snake_to_camel(AzureMLResourceType.ONLINE_ENDPOINT),
            AzureMLResourceType.DEPLOYMENT,
        ):
            endpoint_name, deployment_name = target.endpoint_deployment_id.split(":")  # type: ignore
            target.endpoint_deployment_id = NAMED_RESOURCE_ID_FORMAT_WITH_PARENT.format(
                self._subscription_id,
                self._resource_group_name,
                AZUREML_RESOURCE_PROVIDER,
                self._workspace_name,
                snake_to_camel(AzureMLResourceType.ONLINE_ENDPOINT),
                endpoint_name,
                AzureMLResourceType.DEPLOYMENT,
                deployment_name,
            )
        else:
            deployment_arm_id_entity = AMLNamedArmId(target.endpoint_deployment_id)
            endpoint_name = deployment_arm_id_entity.parent_asset_name
            deployment_name = deployment_arm_id_entity.asset_name

        return endpoint_name, deployment_name
